import unittest
from copy import deepcopy
from typing import Optional, Sequence

import numpy as np
import pandas as pd

from darts.dataprocessing.transformers import Diff
from darts.timeseries import TimeSeries
from darts.timeseries import concatenate as darts_concat
from darts.utils.timeseries_generation import linear_timeseries, sine_timeseries


class DiffTestCase(unittest.TestCase):
    sine_series = [
        5 * sine_timeseries(length=50, value_frequency=f) for f in (0.05, 0.1, 0.15)
    ]
    sine_series = darts_concat(
        sine_series,
        axis=1,
    )

    def assert_series_equal(
        self,
        series1: TimeSeries,
        series2: TimeSeries,
        equal_nan: bool,
        to_compare: Optional[np.ndarray] = None,
    ):

        """
        Helper to compare series differenced by `Diff`.

        Parameters
        ----------
            series1
                First `TimeSeries` object to compare.
            series2
                Second `TimeSeries` object to compare.
            equal_nan
                Whether to compare `NaN` values as equal (see `help(np.testing.assert_allclose)`); should set
                `equal_nan=(not dropna)` (i.e. should throw error if there are `NaN` values when user has requested
                for them to be dropped).
            to_compare
                boolean `np.ndarray` which specifies which subset of columns should be compared in `series1` and
                `series2`. Used when checking that components not specified by `component_mask` have NOT been
                differenced after `fit`ing (i.e. `to_compare = ~component_mask`). If `to_compare = None`, all
                columns in `series1` and `series2` are compared.
        """
        if to_compare is not None:
            series1 = series1.drop_columns(series1.columns[~to_compare])
            series2 = series2.drop_columns(series2.columns[~to_compare])
        np.testing.assert_allclose(
            series1.all_values(), series2.all_values(), atol=1e-8, equal_nan=equal_nan
        )
        self.assertTrue(series1.time_index.equals(series2.time_index))

    def test_diff_quad_series(self):
        """
        Tests that differencing a quadratic series (computed by `cumsum`ing a linear series)
        yields the original linear series.
        """
        lin_series = linear_timeseries(start_value=1, end_value=10, length=50)
        quad_series = TimeSeries(lin_series.data_array().cumsum(axis=0))
        for dropna in (False, True):
            diff = Diff(lags=1, dropna=dropna)
            diff.fit(quad_series)
            # Differencing causes first point to be 'dropped':
            transformed_series = diff.transform(quad_series)
            expected_transform = lin_series.drop_before(0)
            if not dropna:
                expected_transform = expected_transform.prepend_values([np.nan])
            self.assert_series_equal(
                expected_transform, transformed_series, equal_nan=(not dropna)
            )
            self.assert_series_equal(
                series1=quad_series,
                series2=diff.inverse_transform(transformed_series),
                equal_nan=(not dropna),
            )

    def test_diff_inverse_transform_beyond_fit_data(self):
        """
        Tests that `Diff` class can 'undifference' data that extends beyond
         what was used to fit it (e.g. like when undifferencing a forecast of
         a differenced timeseries, so as to yield a forecast of the undifferenced
         time series)
        """

        # (`lag`, `dropna`) pairs:
        test_cases = [
            (1, True),
            (1, False),
            ([1, 2, 3, 2, 1], True),
            ([1, 2, 3, 2, 1], False),
        ]

        # Artifically truncate series:
        short_sine = self.sine_series.copy().drop_after(10)
        for (lags, dropna) in test_cases:
            # Fit Diff to truncated series:
            diff = Diff(lags=lags, dropna=dropna)
            diff.fit(short_sine)
            # Difference entire time series:
            to_undiff = self.sine_series.copy()
            if not isinstance(lags, Sequence):
                lags = (lags,)
            for lag in lags:
                to_undiff = to_undiff.diff(n=1, periods=lag, dropna=dropna)
            # Should be able to undifference entire series even though only fitted
            # to truncated series:
            self.assert_series_equal(
                series1=self.sine_series,
                series2=diff.inverse_transform(to_undiff),
                equal_nan=(not dropna),
            )

    def test_diff_multi_ts(self):
        """
        Tests that `Diff` correctly behaves when given multiple time series, both when a `component_mask`
        is provided and when one is not.
        """

        # (`lags`, `dropna`, `component_mask`) triplets; note that combination of `dropna = True``
        # AND `component_mask is not None` is not tested here, since this should throw error
        # (see `test_diff_dropna_and_component_mask_specified`):
        component_mask = np.array([True, False, True])
        test_cases = [
            (1, True, None),
            (1, False, None),
            ([1, 2, 3, 2, 1], True, None),
            ([1, 2, 3, 2, 1], False, None),
            (1, False, component_mask),
            ([1, 2, 3, 2, 1], False, component_mask),
        ]
        for (lags, dropna, mask) in test_cases:
            diff = Diff(lags=lags, dropna=dropna)
            transformed = diff.fit_transform(
                [self.sine_series, self.sine_series], component_mask=mask
            )
            if mask is not None:
                # Masked components should be undifferenced:
                self.assert_series_equal(
                    self.sine_series,
                    transformed[0],
                    equal_nan=(not dropna),
                    to_compare=~mask,
                )
                self.assert_series_equal(
                    self.sine_series,
                    transformed[1],
                    equal_nan=(not dropna),
                    to_compare=~mask,
                )
            # Should recover original sine_series:
            back = diff.inverse_transform(transformed, component_mask=mask)
            self.assert_series_equal(self.sine_series, back[0], equal_nan=(not dropna))
            self.assert_series_equal(self.sine_series, back[1], equal_nan=(not dropna))

    def test_diff_stochastic_series(self):
        """
        Tests that `Diff` class correctly differences and then undifferences a
        random series with multiple samples.
        """
        test_cases = [
            (1, True),
            (1, False),
            ([1, 2, 3, 2, 1], True),
            ([1, 2, 3, 2, 1], False),
        ]

        vals = np.random.rand(10, 5, 10)
        series = TimeSeries.from_values(vals)

        for (lags, dropna) in test_cases:
            transformer = Diff(lags=lags, dropna=dropna)
            new_series = transformer.fit_transform(series)
            series_back = transformer.inverse_transform(new_series)
            # Should recover original series:
            self.assert_series_equal(series, series_back, equal_nan=(not dropna))

    def test_diff_dropna_and_component_mask_specified(self):
        """
        Tests that `Diff` throws error during `fit` if `component_mask` is specified
        when `dropna = True`; can't allow this since undifferenced components will be different
        length to differenced components.
        """
        diff = Diff(lags=1, dropna=True)
        with self.assertRaises(ValueError) as e:
            diff.fit(self.sine_series, component_mask=np.array([1, 0, 1], dtype=bool))
        self.assertEqual(
            (
                "Cannot specify `component_mask` with `dropna = True`, "
                "since differenced and undifferenced components will be "
                "of different lengths."
            ),
            str(e.exception),
        )

    def test_diff_series_too_short(self):
        """
        Tests that `Diff` throws error is length of series is less than `sum(lags)` (i.e.
        there's not enough data to fit the differencer)
        """
        lags = (1000,)
        diff = Diff(lags=lags)
        with self.assertRaises(ValueError) as e:
            diff.fit(self.sine_series)
        self.assertEqual(
            (
                f"Series requires at least {sum(lags) + 1} timesteps "
                f"to difference with lags {lags}; series only "
                f"has {self.sine_series.n_timesteps} timesteps."
            ),
            str(e.exception),
        )

    def test_diff_incompatible_inverse_transform_date(self):
        """
        Tests that `Diff` throws error when given series to `inverse_transform`
        which starts at a date not equal to the first data of the fitting data.
        """
        vals = np.random.rand(10, 5)
        series1 = TimeSeries.from_times_and_values(
            values=vals, times=pd.date_range(start="1/1/2018", freq="d", periods=10)
        )
        series2 = TimeSeries.from_times_and_values(
            values=vals, times=pd.date_range(start="1/2/2018", freq="d", periods=10)
        )
        for dropna in (False, True):
            diff = Diff(lags=1, dropna=dropna)
            diff.fit(series1)
            series2_diffed = series2.diff(n=1, periods=1, dropna=dropna)
            with self.assertRaises(ValueError) as e:
                diff.inverse_transform(series2_diffed)
            expected_start = (
                series1.start_time()
                if (not dropna)
                else series1.start_time() + series1.freq
            )
            self.assertEqual(
                (
                    f"Expected series to begin at time {expected_start}; "
                    f"instead, it begins at time {series2_diffed.start_time()}."
                ),
                str(e.exception),
            )

    def test_diff_incompatible_inverse_transform_freq(self):
        """
        Tests that `Diff` throws error when given series to `inverse_transform`
        that has different frequency than fitting data.
        """
        vals = np.random.rand(10, 5)
        series1 = TimeSeries.from_times_and_values(
            values=vals, times=pd.date_range(start="1/1/2018", freq="W", periods=10)
        )
        series2 = TimeSeries.from_times_and_values(
            values=vals, times=pd.date_range(start="1/1/2018", freq="M", periods=10)
        )
        diff = Diff(lags=1, dropna=True)
        diff.fit(series1)
        with self.assertRaises(ValueError) as e:
            diff.inverse_transform(series2.diff(n=1, periods=1, dropna=True))
        self.assertEqual(
            f"Series is of frequency {series2.freq}, but transform was fitted to data of frequency {series1.freq}.",
            str(e.exception),
        )

    def test_diff_incompatible_inverse_transform_shape(self):
        """
        Tests that `Diff` throws error when given series to `inverse_transform`
        that has different number of components or samples than the fitting data.
        """
        vals = np.random.rand(10, 5, 5)
        dates = pd.date_range(start="1/1/2018", freq="W", periods=10)
        series = TimeSeries.from_times_and_values(values=vals, times=dates)
        diff = Diff(lags=1, dropna=True)
        diff.fit(series)
        series_rm_comp = TimeSeries.from_times_and_values(
            values=vals[:, 1:, :], times=dates
        )
        with self.assertRaises(ValueError) as e:
            diff.inverse_transform(series_rm_comp.diff(n=1, periods=1, dropna=True))
        self.assertEqual(
            f"Expected series to have {series.n_components} components; "
            f"instead, it has {series.n_components-1}.",
            str(e.exception),
        )
        series_rm_samp = TimeSeries.from_times_and_values(
            values=vals[:, :, 1:], times=dates
        )
        with self.assertRaises(ValueError) as e:
            diff.inverse_transform(series_rm_samp.diff(n=1, periods=1, dropna=True))
        self.assertEqual(
            f"Expected series to have {series.n_samples} samples; "
            f"instead, it has {series.n_samples-1}.",
            str(e.exception),
        )

    def test_diff_multiple_calls_to_fit(self):
        """
        Tests that `Diff` updates `start_vals` parameter when refitted to new data.
        """
        diff = Diff(lags=[2, 2], dropna=True)

        diff.fit(self.sine_series)
        startvals1 = deepcopy(diff._fitted_params)[0][0]

        diff.fit(self.sine_series + 1)
        startvals2 = deepcopy(diff._fitted_params)[0][0]

        self.assertFalse(np.allclose(startvals1, startvals2))



import pandas as pd

from darts.tests.base_test_class import DartsBaseTestClass
from darts.utils.data.tabularization import get_shared_times_bounds
from darts.utils.timeseries_generation import linear_timeseries


class GetSharedTimesBoundsTestCase(DartsBaseTestClass):
    """
    Tests `get_shared_times_bounds` function defined in `darts.utils.data.tabularization`.
    """

    def test_shared_times_bounds_overlapping_range_idx_series(self):
        """
        Tests that `get_shared_times_bounds` correctly computes bounds
        of two overlapping range index timeseries.
        """
        # Defined so `series_1` starts and ends before `series_2` does:
        series_1 = linear_timeseries(start=1, end=15, freq=3)
        series_2 = linear_timeseries(start=2, end=20, freq=2)
        expected_bounds = (series_2.start_time(), series_1.end_time())
        self.assertEqual(
            get_shared_times_bounds(series_1, series_2),
            expected_bounds,
        )

    def test_shared_times_bounds_overlapping_datetime_idx_series(self):
        """
        Tests that `get_shared_times_bounds` correctly computes bounds
        of two overlapping datetime index timeseries.
        """
        # Defined so `series_1` starts and ends before `series_2` does:
        series_1 = linear_timeseries(
            start=pd.Timestamp("1/1/2000"), end=pd.Timestamp("1/15/2000"), freq="3d"
        )
        series_2 = linear_timeseries(
            start=pd.Timestamp("1/2/2000"), end=pd.Timestamp("1/20/2000"), freq="2d"
        )
        expected_bounds = (series_2.start_time(), series_1.end_time())
        self.assertEqual(
            get_shared_times_bounds(series_1, series_2),
            expected_bounds,
        )

    def test_shared_times_bounds_time_idx_inputs(self):
        """
        Tests that `get_shared_times_bounds` behaves correctly
        when passed `pd.Index` inputs instead of `TimeSeries`
        inputs. Mixtures of `pd.Index` and `TimeSeries` inputs
        are also checked.
        """
        # Defined so `series_1` starts and ends before `series_2` does:
        series_1 = linear_timeseries(start=0, end=10, freq=1)
        series_2 = linear_timeseries(start=2, end=16, freq=2)
        expected_bounds = (series_2.start_time(), series_1.end_time())
        # Pass only `series_1.time_index` - bounds are just start and
        # end times of this series:
        self.assertEqual(
            get_shared_times_bounds(series_1.time_index),
            (series_1.start_time(), series_1.end_time()),
        )
        # Pass only `series_2.time_index` - bounds are just start and
        # end times of this series:
        self.assertEqual(
            get_shared_times_bounds(series_2.time_index),
            (series_2.start_time(), series_2.end_time()),
        )
        self.assertEqual(
            get_shared_times_bounds(series_1.time_index, series_2),
            expected_bounds,
        )
        self.assertEqual(
            get_shared_times_bounds(series_1, series_2.time_index),
            expected_bounds,
        )
        self.assertEqual(
            get_shared_times_bounds(series_1.time_index, series_2.time_index),
            expected_bounds,
        )

    def test_shared_times_bounds_subset_series_range_idx(self):
        """
        Tests that `get_shared_times_bounds` correctly handles case where
        the provided series are formed by taking successive subsets of an
        initial series (i.e. `series_2` is formed by taking a subset of
        `series_1`, and `series_3` is formed by taking a subset of `series_2`).
        In such cases, the bounds are simply the start and end times of the
        shortest series. This particular test uses range index series to
        check this behaviour.
        """
        series = linear_timeseries(start=0, length=10, freq=3)
        subseries = (
            series.copy()
            .drop_after(series.time_index[-1])
            .drop_before(series.time_index[1])
        )
        subsubseries = (
            subseries.copy()
            .drop_after(subseries.time_index[-1])
            .drop_before(subseries.time_index[1])
        )
        expected_bounds = (subsubseries.start_time(), subsubseries.end_time())
        self.assertEqual(
            get_shared_times_bounds(series, subseries, subsubseries),
            expected_bounds,
        )

    def test_shared_times_bounds_subset_series_datetime_idx(self):
        """
        Tests that `get_shared_times_bounds` correctly handles case where
        the provided series are formed by taking successive subsets of an
        initial series (i.e. `series_2` is formed by taking a subset of
        `series_1`, and `series_3` is formed by taking a subset of `series_2`).
        In such cases, the bounds are simply the start and end times of the
        shortest series. This particular test uses datetime index series to
        check this behaviour.
        """
        series = linear_timeseries(start=pd.Timestamp("1/1/2000"), length=10, freq="3d")
        subseries = (
            series.copy()
            .drop_after(series.time_index[-1])
            .drop_before(series.time_index[1])
        )
        subsubseries = (
            subseries.copy()
            .drop_after(subseries.time_index[-1])
            .drop_before(subseries.time_index[1])
        )
        expected_bounds = (subsubseries.start_time(), subsubseries.end_time())
        self.assertEqual(
            get_shared_times_bounds(series, subseries, subsubseries),
            expected_bounds,
        )

    def test_shared_times_bounds_identical_inputs_range_idx(self):
        """
        Tests that `get_shared_times_bounds` correctly handles case where
        multiple copies of the same series is passed as an input; we expect
        the return bounds to just be the start and end times of that repeated
        series. This particular test uses range index series to
        check this behaviour.
        """
        series = linear_timeseries(start=0, length=5, freq=1)
        expected = (series.start_time(), series.end_time())
        self.assertEqual(get_shared_times_bounds(series, series), expected)
        self.assertEqual(get_shared_times_bounds(series, series, series), expected)

    def test_shared_times_bounds_identical_inputs_datetime_idx(self):
        """
        Tests that `get_shared_times_bounds` correctly handles case where
        multiple copies of the same series is passed as an input; we expect
        the return bounds to just be the start and end times of that repeated
        series. This particular test uses datetime index series to
        check this behaviour.
        """
        series = linear_timeseries(start=pd.Timestamp("1/1/2000"), length=5, freq="d")
        expected = (series.start_time(), series.end_time())
        self.assertEqual(get_shared_times_bounds(series), expected)
        self.assertEqual(get_shared_times_bounds(series, series), expected)
        self.assertEqual(get_shared_times_bounds(series, series, series), expected)

    def test_shared_times_bounds_unspecified_inputs(self):
        """
        Tests that `get_shared_times_bounds` correctly handles case unspecified
        inputs (i.e. `None`) are passed. If passed with a specified series, the
        `None` input should be ignored, meaning that the returned bounds should
        be the start and end times of the only specified series. If only `None`
        inputs are passed, `None` should be returned.
        """
        # `None` is passed alonside a non-`None` input:
        series = linear_timeseries(start=0, length=5, freq=1)
        expected = (series.start_time(), series.end_time())
        self.assertEqual(get_shared_times_bounds(series, None), expected)
        self.assertEqual(get_shared_times_bounds(None, series), expected)
        self.assertEqual(get_shared_times_bounds(None, series, None), expected)

        # `None` should be returned if no series specified:
        self.assertEqual(get_shared_times_bounds(None), None)
        self.assertEqual(get_shared_times_bounds(None, None, None), None)

    def test_shared_times_bounds_single_idx_overlap_range_idx(self):
        """
        Tests that `get_shared_times_bounds` correctly handles cases
        where the bounds contains a single time index value. This
        particular test uses range time index series to check this
        behaviour.
        """
        # Pass multiple copies of timeseries with single time
        # value - bounds should be start time and end time of
        # this single-valued series:
        series = linear_timeseries(start=0, length=1, freq=1)
        self.assertEqual(
            get_shared_times_bounds(series, series),
            (series.start_time(), series.end_time()),
        )
        # `series_1` and `series_2` share only a single overlap point
        # at the end of `series_1`:
        series_1 = linear_timeseries(start=0, length=3, freq=1)
        series_2 = linear_timeseries(start=series_1.end_time(), length=2, freq=2)
        self.assertEqual(
            get_shared_times_bounds(series_1, series_2),
            (series_1.end_time(), series_2.start_time()),
        )

    def test_shared_times_bounds_single_idx_overlap_datetime_idx(self):
        """
        Tests that `get_shared_times_bounds` correctly handles cases
        where the bounds contains a single time index value. This
        particular test uses range time index series to check this
        behaviour.
        """
        # Pass multiple copies of timeseries with single time
        # value - bounds should be start time and end time of
        # this single-valued series:
        series = linear_timeseries(start=pd.Timestamp("1/1/2000"), length=1, freq="d")
        self.assertEqual(
            get_shared_times_bounds(series, series),
            (series.start_time(), series.end_time()),
        )
        # `series_1` and `series_2` share only a single overlap point
        # at the end of `series_1`:
        series_1 = linear_timeseries(start=pd.Timestamp("1/1/2000"), length=3, freq="d")
        series_2 = linear_timeseries(start=series_1.end_time(), length=2, freq="2d")
        self.assertEqual(
            get_shared_times_bounds(series_1, series_2),
            (series_1.end_time(), series_2.start_time()),
        )

    def test_shared_times_bounds_no_overlap_range_idx(self):
        """
        Tests that `get_shared_times_bounds` returns `None` when provided
        with two series that share no overlap. This particular test uses
        range index series to check this behaviour.
        """
        # Have `series_2` begin after the end of `series_1`:
        series_1 = linear_timeseries(start=0, length=5, freq=1)
        series_2 = linear_timeseries(start=series_1.end_time() + 1, length=6, freq=2)
        self.assertEqual(get_shared_times_bounds(series_1, series_2), None)
        self.assertEqual(get_shared_times_bounds(series_2, series_1, series_2), None)

    def test_shared_times_bounds_no_overlap_datetime_idx(self):
        """
        Tests that `get_shared_times_bounds` returns `None` when provided
        with two series that share no overlap. This particular test uses
        datetime index series to check this behaviour.
        """
        # Have `series_2` begin after the end of `series_1`:
        series_1 = linear_timeseries(start=pd.Timestamp("1/1/2000"), length=5, freq="d")
        series_2 = linear_timeseries(
            start=series_1.end_time() + pd.Timedelta("1d"), length=6, freq="2d"
        )
        self.assertEqual(get_shared_times_bounds(series_1, series_2), None)
        self.assertEqual(get_shared_times_bounds(series_2, series_1, series_2), None)

    def test_shared_times_bounds_different_time_idx_types_error(self):
        """
        Tests that `get_shared_times_bounds` throws correct error
        when a range time index series and a datetime index series
        are specified as inputs together.
        """
        series_1 = linear_timeseries(start=1, length=5, freq=1)
        series_2 = linear_timeseries(start=pd.Timestamp("1/1/2000"), length=5, freq="d")
        with self.assertRaises(ValueError) as e:
            get_shared_times_bounds(series_1, series_2)
        self.assertEqual(
            (
                "Specified series and/or times must all "
                "have the same type of `time_index` (i.e. all "
                "`pd.RangeIndex` or all `pd.DatetimeIndex`)."
            ),
            str(e.exception),
        )

    def test_shared_times_bounds_empty_input(self):
        """
        Tests that `get_shared_times_bounds` returns `None` when
        handed a non-`None` input that has no timesteps.
        """
        series = linear_timeseries(start=0, length=0, freq=1)
        self.assertEqual(get_shared_times_bounds(series), None)
        self.assertEqual(get_shared_times_bounds(series.time_index), None)
        self.assertEqual(get_shared_times_bounds(series, series.time_index), None)



"""
Forecasting Model Base Classes

A forecasting model captures the future values of a time series as a function of the past as follows:

.. math:: y_{t+1} = f(y_t, y_{t-1}, ..., y_1),

where :math:`y_t` represents the time series' value(s) at time :math:`t`.

The main functions are `fit()` and `predict()`. `fit()` learns the function `f()`, over the history of
one or several time series. The function `predict()` applies `f()` on one or several time series in order
to obtain forecasts for a desired number of time stamps into the future.
"""
import copy
import datetime
import inspect
import os
import pickle
import time
from abc import ABC, ABCMeta, abstractmethod
from collections import OrderedDict
from itertools import product
from random import sample
from typing import Any, BinaryIO, Callable, Dict, List, Optional, Sequence, Tuple, Union

import numpy as np
import pandas as pd

from darts import metrics
from darts.dataprocessing.encoders import SequentialEncoder
from darts.logging import get_logger, raise_if, raise_if_not, raise_log
from darts.timeseries import TimeSeries
from darts.utils import (
    _build_tqdm_iterator,
    _historical_forecasts_general_checks,
    _parallel_apply,
    _with_sanity_checks,
)
from darts.utils.timeseries_generation import (
    _build_forecast_series,
    _generate_new_dates,
    generate_index,
)
from darts.utils.utils import (
    drop_after_index,
    drop_before_index,
    get_single_series,
    series2seq,
)

logger = get_logger(__name__)


class ModelMeta(ABCMeta):
    """Meta class to store parameters used at model creation.

    When creating a model instance, the parameters are extracted as follows:

        1)  Get the model's __init__ signature and store all arg and kwarg
            names as well as default values (empty for args) in an ordered
            dict `all_params`.
        2)  Replace the arg values from `all_params` with the positional
            args used at model creation.
        3)  Remove args from `all_params` that were not passed as positional
            args at model creation. This will enforce that an error is raised
            if not all positional args were passed. If all positional args
            were passed, no parameter will be removed.
        4)  Update `all_params` kwargs with optional kwargs from model creation.
        5)  Save `all_params` to the model.
        6)  Call (create) the model with `all_params`.
    """

    def __call__(cls, *args, **kwargs):
        # 1) get all default values from class' __init__ signature
        sig = inspect.signature(cls.__init__)
        all_params = OrderedDict(
            [
                (p.name, p.default)
                for p in sig.parameters.values()
                if not p.name == "self"
            ]
        )

        # 2) fill params with positional args
        for param, arg in zip(all_params, args):
            all_params[param] = arg

        # 3) remove args which were not set (and are per default empty)
        remove_params = []
        for param, val in all_params.items():
            if val is sig.parameters[param].empty:
                remove_params.append(param)
        for param in remove_params:
            all_params.pop(param)

        # 4) update defaults with actual model call parameters and store
        all_params.update(kwargs)

        # 5) save parameters in model
        cls._model_call = all_params

        # 6) call model
        return super().__call__(**all_params)


class ForecastingModel(ABC, metaclass=ModelMeta):
    """The base class for forecasting models. It defines the *minimal* behavior that all forecasting models have to
    support. The signatures in this base class are for "local" models handling only one univariate series and no
    covariates. Sub-classes can handle more complex cases.
    """

    @abstractmethod
    def __init__(self, *args, **kwargs):
        # The series, and covariates used for training the model through the `fit()` function.
        # This is only used if the model has been fit on one time series.
        self.training_series: Optional[TimeSeries] = None
        self.past_covariate_series: Optional[TimeSeries] = None
        self.future_covariate_series: Optional[TimeSeries] = None
        self.static_covariates: Optional[pd.DataFrame] = None

        self._expect_past_covariates, self._uses_past_covariates = False, False
        self._expect_future_covariates, self._uses_future_covariates = False, False
        # for static covariates there is the option to consider static covariates or ignore them
        self._considers_static_covariates = False
        self._expect_static_covariates, self._uses_static_covariates = False, False

        # state; whether the model has been fit (on a single time series)
        self._fit_called = False

        # extract and store sub class model creation parameters
        self._model_params = self._extract_model_creation_params()

        if "add_encoders" not in kwargs:
            raise_log(
                NotImplementedError(
                    "Model subclass must pass the `add_encoders` parameter to base class."
                ),
                logger=logger,
            )

        # by default models do not use encoders
        self.add_encoders = kwargs["add_encoders"]
        self.encoders: Optional[SequentialEncoder] = None

    @abstractmethod
    def fit(self, series: TimeSeries) -> "ForecastingModel":
        """Fit/train the model on the provided series.

        Parameters
        ----------
        series
            A target time series. The model will be trained to forecast this time series.

        Returns
        -------
        self
            Fitted model.
        """
        raise_if_not(
            len(series) >= self.min_train_series_length,
            "Train series only contains {} elements but {} model requires at least {} entries".format(
                len(series), str(self), self.min_train_series_length
            ),
        )
        self.training_series = series
        self._fit_called = True

        if series.has_range_index:
            self._supports_range_index()

    def _supports_range_index(self) -> bool:
        """Checks if the forecasting model supports a range index.
        Some models may not support this, if for instance they rely on underlying dates.

        By default, returns True. Needs to be overwritten by models that do not support
        range indexing and raise meaningful exception.
        """
        return True

    def _is_probabilistic(self) -> bool:
        """
        Checks if the forecasting model supports probabilistic predictions.
        By default, returns False. Needs to be overwritten by models that do support
        probabilistic predictions.
        """
        return False

    def _supports_non_retrainable_historical_forecasts(self) -> bool:
        """
        Checks if the forecasting model supports historical forecasts without retraining
        the model. By default, returns False. Needs to be overwritten by models that do
        support historical forecasts without retraining.
        """
        return False

    @property
    def supports_past_covariates(self):
        """
        Whether model supports past covariates
        """
        return "past_covariates" in inspect.signature(self.fit).parameters.keys()

    @property
    def supports_future_covariates(self):
        """
        Whether model supports future covariates
        """
        return "future_covariates" in inspect.signature(self.fit).parameters.keys()

    @property
    def supports_static_covariates(self) -> bool:
        """
        Whether model supports static covariates
        """
        return False

    @property
    def uses_past_covariates(self):
        """
        Whether the model uses past covariates, once fitted.
        """
        return self._uses_past_covariates

    @property
    def uses_future_covariates(self):
        """
        Whether the model uses future covariates, once fitted.
        """
        return self._uses_future_covariates

    @property
    def uses_static_covariates(self):
        """
        Whether the model uses static covariates, once fitted.
        """
        return self._uses_static_covariates

    @property
    def considers_static_covariates(self):
        """
        Whether the model considers static covariates, if there are any.
        """
        return self._considers_static_covariates

    @abstractmethod
    def predict(self, n: int, num_samples: int = 1) -> TimeSeries:
        """Forecasts values for `n` time steps after the end of the training series.

        Parameters
        ----------
        n
            Forecast horizon - the number of time steps after the end of the series for which to produce predictions.
        num_samples
            Number of times a prediction is sampled from a probabilistic model. Should be left set to 1
            for deterministic models.

        Returns
        -------
        TimeSeries
            A time series containing the `n` next points after then end of the training series.
        """
        if not self._fit_called:
            raise_log(
                ValueError(
                    "The model must be fit before calling predict(). "
                    "For global models, if predict() is called without specifying a series, "
                    "the model must have been fit on a single training series."
                ),
                logger,
            )

        if not self._is_probabilistic() and num_samples > 1:
            raise_log(
                ValueError(
                    "`num_samples > 1` is only supported for probabilistic models."
                ),
                logger,
            )

    def _fit_wrapper(
        self,
        series: TimeSeries,
        past_covariates: Optional[TimeSeries],
        future_covariates: Optional[TimeSeries],
    ):
        self.fit(series)

    def _predict_wrapper(
        self,
        n: int,
        series: TimeSeries,
        past_covariates: Optional[TimeSeries],
        future_covariates: Optional[TimeSeries],
        num_samples: int,
        verbose: bool = False,
    ) -> TimeSeries:
        return self.predict(n, num_samples=num_samples, verbose=verbose)

    @property
    def min_train_series_length(self) -> int:
        """
        The minimum required length for the training series.
        """
        return 3

    @property
    def min_train_samples(self) -> int:
        """
        The minimum number of samples for training the model.
        """
        return 1

    @property
    @abstractmethod
    def extreme_lags(
        self,
    ) -> Tuple[
        Optional[int],
        Optional[int],
        Optional[int],
        Optional[int],
        Optional[int],
        Optional[int],
    ]:
        """
        A 6-tuple containing in order:
        (min target lag, max target lag, min past covariate lag, max past covariate lag, min future covariate
        lag, max future covariate lag). If 0 is the index of the first prediction, then all lags are relative to this
        index.

        See examples below.

        If the model wasn't fitted with:
            - target (concerning RegressionModels only): then the first element should be `None`.

            - past covariates: then the third and fourth elements should be `None`.

            - future covariates: then the fifth and sixth elements should be `None`.

        Should be overridden by models that use past or future covariates, and/or for model that have minimum target
        lag and maximum target lags potentially different from -1 and 0.

        Notes
        -----
        maximum target lag (second value) cannot be `None` and is always larger than or equal to 0.
        Examples
        --------
        >>> model = LinearRegressionModel(lags=3, output_chunk_length=2)
        >>> model.fit(train_series)
        >>> model.extreme_lags
        (-3, 1, None, None, None, None)
        >>> model = LinearRegressionModel(lags=[-3, -5], lags_past_covariates = 4, output_chunk_length=7)
        >>> model.fit(train_series, past_covariates=past_covariates)
        >>> model.extreme_lags
        (-5, 6, -4, -1,  None, None)
        >>> model = LinearRegressionModel(lags=[3, 5], lags_future_covariates = [4, 6], output_chunk_length=7)
        >>> model.fit(train_series, future_covariates=future_covariates)
        >>> model.extreme_lags
        (-5, 6, None, None, 4, 6)
        >>> model = NBEATSModel(input_chunk_length=10, output_chunk_length=7)
        >>> model.fit(train_series)
        >>> model.extreme_lags
        (-10, 6, None, None, None, None)
        >>> model = NBEATSModel(input_chunk_length=10, output_chunk_length=7, lags_future_covariates=[4, 6])
        >>> model.fit(train_series, future_covariates)
        >>> model.extreme_lags
        (-10, 6, None, None, 4, 6)
        """
        pass

    @property
    def _training_sample_time_index_length(self) -> int:
        """
        Required time_index length for one training sample, for any model.
        """
        (
            min_target_lag,
            max_target_lag,
            min_past_cov_lag,
            max_past_cov_lag,
            min_future_cov_lag,
            max_future_cov_lag,
        ) = self.extreme_lags

        return max(
            max_target_lag + 1,
            max_future_cov_lag + 1 if max_future_cov_lag else 0,
        ) - min(
            min_target_lag if min_target_lag else 0,
            min_past_cov_lag if min_past_cov_lag else 0,
            min_future_cov_lag if min_future_cov_lag else 0,
        )

    @property
    def _predict_sample_time_index_length(self) -> int:
        """
        Required time_index length for one `predict` function call, for any model.
        """
        (
            min_target_lag,
            max_target_lag,
            min_past_cov_lag,
            max_past_cov_lag,
            min_future_cov_lag,
            max_future_cov_lag,
        ) = self.extreme_lags

        return (max_future_cov_lag + 1 if max_future_cov_lag else 0) - min(
            min_target_lag if min_target_lag else 0,
            min_past_cov_lag if min_past_cov_lag else 0,
            min_future_cov_lag if min_future_cov_lag else 0,
        )

    @property
    def _predict_sample_time_index_past_length(self) -> int:
        """
        Required time_index length in the past for one `predict` function call, for any model.
        """
        (
            min_target_lag,
            max_target_lag,
            min_past_cov_lag,
            max_past_cov_lag,
            min_future_cov_lag,
            max_future_cov_lag,
        ) = self.extreme_lags

        return -min(
            min_target_lag if min_target_lag else 0,
            min_past_cov_lag if min_past_cov_lag else 0,
            min_future_cov_lag if min_future_cov_lag else 0,
        )

    def _get_historical_forecastable_time_index(
        self,
        series: TimeSeries,
        past_covariates: Optional[TimeSeries] = None,
        future_covariates: Optional[TimeSeries] = None,
        is_training: Optional[bool] = False,
    ) -> Union[pd.DatetimeIndex, pd.RangeIndex, None]:
        """
        Private function that returns the largest time_index representing the subset of each timestamps
        for which historical forecasts can be made, given the model's properties, the training series
        and the covariates.
            - If ``None`` is returned, there is no point where a forecast can be made.

            - If ``is_training=False``, it returns the time_index subset of predictable timestamps.

            - If ``is_training=True``, it returns the time_index subset of trainable timestamps. A trainable
            timestamp is a timestamp that has a training sample of length at least ``self.training_sample_length``
             preceding it.


        Parameters
        ----------
        series
            A target series.
        past_covariates
            Optionally, a past covariates.
        future_covariates
            Optionally, a future covariates.
        is_training
            Whether the returned time_index should be taking into account the training.

        Returns
        -------
        Union[pd.DatetimeIndex, pd.RangeIndex, None]
            The longest time_index that can be used for historical forecasting.

        Examples
        --------
        >>> model = LinearRegressionModel(lags=3, output_chunk_length=2)
        >>> model.fit(train_series)
        >>> series = TimeSeries.from_times_and_values(pd.date_range('2000-01-01', '2000-01-10'), np.arange(10))
        >>> model._get_historical_forecastable_time_index(series=series, is_training=False)
        DatetimeIndex(['2000-01-04', '2000-01-05', '2000-01-06', '2000-01-07',
               '2000-01-08', '2000-01-09', '2000-01-10'],
              dtype='datetime64[ns]', freq='D')
        >>> model._get_historical_forecastable_time_index(series=series, is_training=True)
        DatetimeIndex(['2000-01-06', '2000-01-08', '2000-01-09', '2000-01-10'],
                dtype='datetime64[ns]', freq='D')

        >>> model = NBEATSModel(input_chunk_length=3, output_chunk_length=3)
        >>> model.fit(train_series, train_past_covariates)
        >>> series = TimeSeries.from_times_and_values(pd.date_range('2000-10-01', '2000-10-09'), np.arange(8))
        >>> past_covariates = TimeSeries.from_times_and_values(pd.date_range('2000-10-03', '2000-10-20'),
        np.arange(18))
        >>> model._get_historical_forecastable_time_index(series=series, past_covariates=past_covariates,
        is_training=False)
        DatetimeIndex(['2000-10-06', '2000-10-07', '2000-10-08', '2000-10-09'], dtype='datetime64[ns]', freq='D')
        >>> model._get_historical_forecastable_time_index(series=series, past_covariates=past_covariates,
        is_training=True)
        DatetimeIndex(['2000-10-09'], dtype='datetime64[ns]', freq='D') # Only one point is trainable, and
        # corresponds to the first point after we reach a common subset of timestamps of training_sample_length length.
        """

        (
            min_target_lag,
            max_target_lag,
            min_past_cov_lag,
            max_past_cov_lag,
            min_future_cov_lag,
            max_future_cov_lag,
        ) = self.extreme_lags

        if min_target_lag is None:
            min_target_lag = 0

        # longest possible time index for target
        intersect_ = generate_index(
            start=series.start_time()
            + (max_target_lag - min_target_lag + 1) * series.freq
            if is_training
            else series.start_time() - min_target_lag * series.freq,
            end=series.end_time() + 1 * series.freq,
            freq=series.freq,
        )

        # longest possible time index for past covariates
        if (min_past_cov_lag is not None) and (past_covariates is not None):
            tmp_ = generate_index(
                start=past_covariates.start_time()
                - (min_past_cov_lag - max_target_lag - 1) * past_covariates.freq
                if is_training
                else past_covariates.start_time()
                - min_past_cov_lag * past_covariates.freq,
                end=past_covariates.end_time()
                - max_past_cov_lag * past_covariates.freq,
                freq=past_covariates.freq,
            )
            if intersect_ is not None:
                intersect_ = intersect_.intersection(tmp_)
            else:
                intersect_ = tmp_

        # longest possible time index for future covariates
        if (min_future_cov_lag is not None) and (future_covariates is not None):
            tmp_ = generate_index(
                start=future_covariates.start_time()
                - (min_future_cov_lag - max_target_lag - 1) * future_covariates.freq
                if is_training
                else future_covariates.start_time()
                - min_future_cov_lag * future_covariates.freq,
                end=future_covariates.end_time()
                - max_future_cov_lag * future_covariates.freq,
                freq=future_covariates.freq,
            )

            if intersect_ is not None:
                intersect_ = intersect_.intersection(tmp_)
            else:
                intersect_ = tmp_

        return intersect_ if len(intersect_) > 0 else None

    def _generate_new_dates(
        self, n: int, input_series: Optional[TimeSeries] = None
    ) -> Union[pd.DatetimeIndex, pd.RangeIndex]:
        """
        Generates `n` new dates after the end of the specified series
        """
        input_series = (
            input_series if input_series is not None else self.training_series
        )
        return _generate_new_dates(n=n, input_series=input_series)

    def _build_forecast_series(
        self,
        points_preds: Union[np.ndarray, Sequence[np.ndarray]],
        input_series: Optional[TimeSeries] = None,
    ) -> TimeSeries:
        """
        Builds a forecast time series starting after the end of the training time series, with the
        correct time index (or after the end of the input series, if specified).
        """
        input_series = (
            input_series if input_series is not None else self.training_series
        )
        return _build_forecast_series(points_preds, input_series)

    def _historical_forecasts_sanity_checks(self, *args: Any, **kwargs: Any) -> None:
        """Sanity checks for the historical_forecasts function

        Parameters
        ----------
        args
            The args parameter(s) provided to the historical_forecasts function.
        kwargs
            The kwargs parameter(s) provided to the historical_forecasts function.

        Raises
        ------
        ValueError
            when a check on the parameter does not pass.
        """
        # parse args and kwargs
        series = args[0]
        _historical_forecasts_general_checks(series, kwargs)

    def _get_last_prediction_time(self, series, forecast_horizon, overlap_end):
        if overlap_end:
            last_valid_pred_time = series.time_index[-1]
        else:
            last_valid_pred_time = series.time_index[-forecast_horizon]

        return last_valid_pred_time

    @_with_sanity_checks("_historical_forecasts_sanity_checks")
    def historical_forecasts(
        self,
        series: Union[TimeSeries, Sequence[TimeSeries]],
        past_covariates: Optional[Union[TimeSeries, Sequence[TimeSeries]]] = None,
        future_covariates: Optional[Union[TimeSeries, Sequence[TimeSeries]]] = None,
        num_samples: int = 1,
        train_length: Optional[int] = None,
        start: Optional[Union[pd.Timestamp, float, int]] = None,
        forecast_horizon: int = 1,
        stride: int = 1,
        retrain: Union[bool, int, Callable[..., bool]] = True,
        overlap_end: bool = False,
        last_points_only: bool = True,
        verbose: bool = False,
        show_warnings: bool = True,
    ) -> Union[
        TimeSeries, List[TimeSeries], Sequence[TimeSeries], Sequence[List[TimeSeries]]
    ]:
        """Compute the historical forecasts that would have been obtained by this model on
        (potentially multiple) `series`.

        This method repeatedly builds a training set: either expanding from the beginning of `series` or moving with
        a fixed length `train_length`. It trains the model on the training set, emits a forecast of length equal to
        forecast_horizon, and then moves the end of the training set forward by `stride` time steps.

        By default, this method will return one (or a sequence of) single time series made up of
        the last point of each historical forecast.
        This time series will thus have a frequency of ``series.freq * stride``.
        If `last_points_only` is set to False, it will instead return one (or a sequence of) list of the
        historical forecasts series.

        By default, this method always re-trains the models on the entire available history, corresponding to an
        expanding window strategy. If `retrain` is set to False, the model must have been fit before. This is not
        supported by all models.

        Parameters
        ----------
        series
            The (or a sequence of) target time series used to successively train and compute the historical forecasts.
        past_covariates
            Optionally, one (or a sequence of) past-observed covariate series. This applies only if the model
            supports past covariates.
        future_covariates
            Optionally, one (or a sequence of) of future-known covariate series. This applies only if the model
            supports future covariates.
        num_samples
            Number of times a prediction is sampled from a probabilistic model. Use values `>1` only for probabilistic
            models.
        train_length
            Number of time steps in our training set (size of backtesting window to train on). Only effective when
            `retrain` is not ``False``. Default is set to `train_length=None` where it takes all available time steps
            up until prediction time, otherwise the moving window strategy is used. If larger than the number of time
            steps available, all steps up until prediction time are used, as in default case. Needs to be at least
            `min_train_series_length`.
        start
            Optionally, the first point in time at which a prediction is computed for a future time.
            This parameter supports: ``float``, ``int`` and ``pandas.Timestamp``, and ``None``.
            If a ``float``, the parameter will be treated as the proportion of the time series
            that should lie before the first prediction point.
            If an ``int``, the parameter will be treated as an integer index to the time index of
            `series` that will be used as first prediction time.
            If a ``pandas.Timestamp``, the time stamp will be used to determine the first prediction time
            directly.
            If ``None``, the first prediction time will automatically be set to:
                 - the first predictable point if `retrain` is ``False``, or `retrain` is a Callable and the first
                 predictable point is earlier than the first trainable point.

                 - the first trainable point if `retrain` is ``True`` or ``int`` (given `train_length`),
                 or `retrain` is a Callable and the first trainable point is earlier than the first predictable point.

                 - the first trainable point (given `train_length`) otherwise
            Note: Raises a ValueError if `start` yields a time outside the time index of `series`.
            Note: If `start` is outside the possible historical forecasting times, will ignore the parameter
            (default behavior with ``None``) and start at the first trainable/predictable point.
        forecast_horizon
            The forecast horizon for the predictions.
        stride
            The number of time steps between two consecutive predictions.
        retrain
            Whether and/or on which condition to retrain the model before predicting.
            This parameter supports 3 different datatypes: ``bool``, (positive) ``int``, and
            ``Callable`` (returning a ``bool``).
            In the case of ``bool``: retrain the model at each step (`True`), or never retrains the model (`False`).
            In the case of ``int``: the model is retrained every `retrain` iterations.
            In the case of ``Callable``: the model is retrained whenever callable returns `True`.
            The callable must have the following positional arguments:

                - `counter` (int): current `retrain` iteration
                - `pred_time` (pd.Timestamp or int): timestamp of forecast time (end of the training series)
                - `train_series` (TimeSeries): train series up to `pred_time`
                - `past_covariates` (TimeSeries): past_covariates series up to `pred_time`
                - `future_covariates` (TimeSeries): future_covariates series up
                  to `min(pred_time + series.freq * forecast_horizon, series.end_time())`

            Note: if any optional `*_covariates` are not passed to `historical_forecast`, ``None`` will be passed
            to the corresponding retrain function argument.
            Note: some models do require being retrained every time and do not support anything other
            than `retrain=True`.
        overlap_end
            Whether the returned forecasts can go beyond the series' end or not.
        last_points_only
            Whether to retain only the last point of each historical forecast.
            If set to True, the method returns a single ``TimeSeries`` containing the successive point forecasts.
            Otherwise, returns a list of historical ``TimeSeries`` forecasts.
        verbose
            Whether to print progress.
        show_warnings
            Whether to show warnings related to parameters `start`, and `train_length`.

        Returns
        -------
        TimeSeries or List[TimeSeries] or List[List[TimeSeries]]
            If `last_points_only` is set to True and a single series is provided in input,
            a single ``TimeSeries`` is returned, which contains the the historical forecast
            at the desired horizon.

            A ``List[TimeSeries]`` is returned if either `series` is a ``Sequence`` of ``TimeSeries``,
            or if `last_points_only` is set to False. A list of lists is returned if both conditions are met.
            In this last case, the outer list is over the series provided in the input sequence,
            and the inner lists contain the different historical forecasts.
        """
        model: ForecastingModel = self
        # only GlobalForecastingModels support historical forecasting without retraining the model
        base_class_name = model.__class__.__base__.__name__

        # we can directly abort if retrain is False and fit hasn't been called before
        raise_if(
            not model._fit_called and retrain is False,
            "The model has not been fitted yet, and `retrain` is ``False``. "
            "Either call `fit()` before `historical_forecasts()`, or set `retrain` "
            "to something different than ``False``.",
            logger,
        )

        raise_if(
            (isinstance(retrain, Callable) or int(retrain) != 1)
            and (not model._supports_non_retrainable_historical_forecasts()),
            f"{base_class_name} does not support historical forecasting with `retrain` set to `False`. "
            f"For now, this is only supported with GlobalForecastingModels such as TorchForecastingModels. "
            f"For more information, read the documentation for `retrain` in `historical_forecasts()`",
            logger,
        )

        if train_length and not isinstance(train_length, int):
            raise_log(
                TypeError("If not None, train_length needs to be an integer."),
                logger,
            )
        elif (train_length is not None) and train_length < 1:
            raise_log(
                ValueError("If not None, train_length needs to be positive."),
                logger,
            )
        elif (
            train_length is not None
        ) and train_length < model._training_sample_time_index_length + (
            self.min_train_samples - 1
        ):
            raise_log(
                ValueError(
                    "train_length is too small for the training requirements of this model. "
                    f"Must be `>={model._training_sample_time_index_length + (self.min_train_samples - 1)}`."
                ),
                logger,
            )
        if train_length is not None and retrain is False:
            raise_log(
                ValueError("cannot use `train_length` when `retrain=False`."),
                logger,
            )

        if isinstance(retrain, bool) or (isinstance(retrain, int) and retrain >= 0):

            def retrain_func(
                counter, pred_time, train_series, past_covariates, future_covariates
            ):
                return counter % int(retrain) == 0 if retrain else False

        elif isinstance(retrain, Callable):
            retrain_func = retrain

            # check that the signature matches the documentation
            expected_arguments = [
                "counter",
                "pred_time",
                "train_series",
                "past_covariates",
                "future_covariates",
            ]
            passed_arguments = list(inspect.signature(retrain_func).parameters.keys())
            raise_if(
                expected_arguments != passed_arguments,
                f"the Callable `retrain` must have a signature/arguments matching the following positional arguments: "
                f"{expected_arguments}.",
                logger,
            )

            # passing dummy values to check the type of the output
            result = retrain_func(
                counter=0,
                pred_time=get_single_series(series).time_index[-1],
                train_series=series,
                past_covariates=past_covariates,
                future_covariates=future_covariates,
            )
            raise_if_not(
                isinstance(result, bool),
                f"Return value of `retrain` must be bool, received {type(result)}",
                logger,
            )
        else:
            retrain_func = None
            raise_log(
                ValueError(
                    "`retrain` argument must be either `bool`, positive `int` or `Callable` (returning `bool`)"
                ),
                logger,
            )

        series = series2seq(series)
        past_covariates = series2seq(past_covariates)
        future_covariates = series2seq(future_covariates)

        if len(series) == 1:
            # Use tqdm on the outer loop only if there's more than one series to iterate over
            # (otherwise use tqdm on the inner loop).
            outer_iterator = series
        else:
            outer_iterator = _build_tqdm_iterator(series, verbose)

        forecasts_list = []
        for idx, series_ in enumerate(outer_iterator):
            train_length_ = None
            past_covariates_ = past_covariates[idx] if past_covariates else None
            future_covariates_ = future_covariates[idx] if future_covariates else None

            # Prediction
            historical_forecasts_time_index_predict = (
                model._get_historical_forecastable_time_index(
                    series_,
                    past_covariates_,
                    future_covariates_,
                    is_training=False,
                )
            )

            if historical_forecasts_time_index_predict is None:
                raise_log(
                    ValueError(
                        "Cannot build a single input for prediction with the provided model, "
                        f"`series` and `*_covariates` at series index: {idx}. The minimum "
                        "prediction input time index requirements were not met. "
                        "Please check the time index of `series` and `*_covariates`."
                    ),
                    logger,
                )

            # Train
            historical_forecasts_time_index_train = (
                model._get_historical_forecastable_time_index(
                    series_,
                    past_covariates_,
                    future_covariates_,
                    is_training=True,
                )
            )

            if (
                (retrain is not False) or (not model._fit_called)
            ) and historical_forecasts_time_index_train is None:
                raise_log(
                    ValueError(
                        "Cannot build a single input for training with the provided untrained model, "
                        f"`series` and `*_covariates` at series index: {idx}. The minimum "
                        "training input time index requirements were not met. "
                        "Please check the time index of `series` and `*_covariates`."
                    ),
                    logger,
                )

            # We need the first value timestamp to be used in order to properly shift the series
            if historical_forecasts_time_index_train is None:
                # dummy value, should not impact the downstream operations
                min_timestamp_train = series_.time_index[0]
            else:
                # look at both past and future, since the target lags must be taken in consideration
                min_timestamp_train = (
                    historical_forecasts_time_index_train[0]
                    - model._training_sample_time_index_length * series_.freq
                )

            if isinstance(retrain, Callable):
                # retain the longer time index, anything can happen
                if (
                    historical_forecasts_time_index_train is not None
                    and historical_forecasts_time_index_train[0]
                    < historical_forecasts_time_index_predict[0]
                ):
                    historical_forecasts_time_index = (
                        historical_forecasts_time_index_train
                    )
                else:
                    historical_forecasts_time_index = (
                        historical_forecasts_time_index_predict
                    )
            elif retrain:
                historical_forecasts_time_index = historical_forecasts_time_index_train
            else:
                historical_forecasts_time_index = (
                    historical_forecasts_time_index_predict
                )

            # compute the maximum forecasts time index assuming that `start=None`
            if retrain or (not model._fit_called):
                if train_length and train_length <= len(series_):
                    train_length_ = train_length
                    # we have to start later for larger `train_length`
                    step_ahead = max(
                        train_length - model._training_sample_time_index_length, 0
                    )
                    if step_ahead:
                        historical_forecasts_time_index = drop_before_index(
                            historical_forecasts_time_index,
                            historical_forecasts_time_index[0]
                            + step_ahead * series_.freq,
                        )
                # if not we start training right away; some models (sklearn) require more than 1
                # training samples, so we start after the first trainable point.
                else:
                    if train_length and train_length > len(series_) and show_warnings:
                        logger.warning(
                            f"`train_length` is larger than the length of series at index: {idx}. "
                            f"Ignoring `train_length` and using default behavior where all available time steps up "
                            f"until the end of the expanding training set."
                        )

                    train_length_ = None
                    if self.min_train_samples > 1:
                        historical_forecasts_time_index = drop_before_index(
                            historical_forecasts_time_index,
                            historical_forecasts_time_index[0]
                            + (self.min_train_samples - 1) * series_.freq,
                        )

            # Take into account overlap_end, and forecast_horizon.
            last_valid_pred_time = model._get_last_prediction_time(
                series_,
                forecast_horizon,
                overlap_end,
            )

            # The historical_forecasts_time_index end (which was just model dependent so far) is readjusted
            # by function parameters overlap_end and forecast_horizon.
            historical_forecasts_time_index = drop_after_index(
                historical_forecasts_time_index, last_valid_pred_time
            )

            # adjust maximum index with optional `start` value
            if start is not None:
                start_time_ = series_.get_timestamp_at_point(start)
                if (
                    not historical_forecasts_time_index[0]
                    <= start_time_
                    <= historical_forecasts_time_index[-1]
                ):
                    if show_warnings:
                        if not isinstance(start, pd.Timestamp):
                            start_value_msg = f"value `{start}` corresponding to timestamp `{start_time_}`"
                        else:
                            start_value_msg = f"time `{start_time_}`"

                        if start_time_ < historical_forecasts_time_index[0]:
                            logger.warning(
                                f"`start` {start_value_msg} is before the first predictable/trainable historical "
                                f"forecasting point for series at index: {idx}. Ignoring `start` for this series and "
                                f"beginning at first trainable/predictable time: {historical_forecasts_time_index[0]}. "
                                f"To hide these warnings, set `show_warnings=False`."
                            )
                        else:
                            logger.warning(
                                f"`start` {start_value_msg} is after the last trainable/predictable historical "
                                f"forecasting point for series at index: {idx}. This would results in empty historical "
                                f"forecasts. Ignoring `start` for this series and beginning at first trainable/"
                                f"predictable time: {historical_forecasts_time_index[0]}. Non-empty forecasts can be "
                                f"generated by setting `start` value to times between (including): "
                                f"{historical_forecasts_time_index[0], historical_forecasts_time_index[-1]}. "
                                f"To hide these warnings, set `show_warnings=False`."
                            )
                    # ignore user-defined `start`
                    start_time_ = None

                if start_time_ is not None:
                    historical_forecasts_time_index = drop_before_index(
                        historical_forecasts_time_index,
                        start_time_,
                    )

            if len(series) == 1:
                # Only use tqdm if there's no outer loop
                iterator = _build_tqdm_iterator(
                    historical_forecasts_time_index[::stride], verbose
                )
            else:
                iterator = historical_forecasts_time_index[::stride]

            # Either store the whole forecasts or only the last points of each forecast, depending on last_points_only
            forecasts = []

            last_points_times = []
            last_points_values = []
            _counter_train = 0

            # iterate and forecast
            for _counter, pred_time in enumerate(iterator):
                # build the training series, as if retrain is True to break circular dependency
                if min_timestamp_train > series_.time_index[0]:
                    train_series = series_.drop_before(
                        min_timestamp_train - 1 * series_.freq
                    ).drop_after(pred_time)
                else:
                    train_series = series_.drop_after(pred_time)

                if train_length_ and len(train_series) > train_length_:
                    train_series = train_series[-train_length_:]

                # testing `retrain` to exclude `False` and `0`
                if (
                    retrain
                    and historical_forecasts_time_index_train is not None
                    and pred_time in historical_forecasts_time_index_train
                ):
                    # retrain_func processes the series that would be used for training
                    if retrain_func(
                        counter=_counter_train,
                        pred_time=pred_time,
                        train_series=train_series,
                        past_covariates=past_covariates_,
                        future_covariates=future_covariates_,
                    ):
                        # avoid fitting the same model multiple times
                        model = model.untrained_model()
                        model._fit_wrapper(
                            series=train_series,
                            past_covariates=past_covariates_,
                            future_covariates=future_covariates_,
                        )
                    else:
                        # untrained model was not trained on the first trainable timestamp
                        if not _counter_train and not model._fit_called:
                            raise_log(
                                ValueError(
                                    f"`retrain` is `False` in the first train iteration at prediction point (in time) "
                                    f"`{pred_time}` and the model has not been fit before. Either call `fit()` before "
                                    f"`historical_forecasts()`, use a different `retrain` value or modify the function "
                                    f"to return `True` at or before this timestamp."
                                ),
                                logger,
                            )
                    _counter_train += 1
                else:
                    # model must be fit before the first prediction
                    # `historical_forecasts_time_index_train` is known to be not None
                    if not _counter and not model._fit_called:
                        raise_log(
                            ValueError(
                                f"Model has not been fit before the first predict iteration at prediction point "
                                f"(in time) `{pred_time}`. Either call `fit()` before `historical_forecasts()`, "
                                f"set `retrain=True`, modify the function to return `True` at least once before "
                                f"`{pred_time}`, or use a different `start` value. The first 'predictable' "
                                f"timestamp with re-training inside `historical_forecasts` is: "
                                f"{historical_forecasts_time_index_train[0]} (potential `start` value)."
                            ),
                            logger,
                        )
                    # slice the series for prediction without retraining
                    train_series = series_.drop_after(pred_time)

                # for regression models with lags=None, lags_past_covariates=None and min(lags_future_covariates)>=0,
                # the first predictable timestamp is the first timestamp of the series, a dummy ts must be created
                # to support `predict()`
                if len(train_series) == 0:
                    train_series = TimeSeries.from_times_and_values(
                        times=generate_index(
                            start=pred_time - 1 * series_.freq,
                            length=1,
                            freq=series_.freq,
                        ),
                        values=np.array([np.NaN]),
                    )

                forecast = model._predict_wrapper(
                    n=forecast_horizon,
                    series=train_series,
                    past_covariates=past_covariates_,
                    future_covariates=future_covariates_,
                    num_samples=num_samples,
                    verbose=verbose,
                )

                if last_points_only:
                    last_points_values.append(forecast.all_values(copy=False)[-1])
                    last_points_times.append(forecast.end_time())
                else:
                    forecasts.append(forecast)

            if last_points_only:
                forecasts_list.append(
                    TimeSeries.from_times_and_values(
                        generate_index(
                            start=last_points_times[0],
                            end=last_points_times[-1],
                            freq=series_.freq * stride,
                        ),
                        np.array(last_points_values),
                        columns=series_.columns,
                        static_covariates=series_.static_covariates,
                        hierarchy=series_.hierarchy,
                    )
                )
            else:
                forecasts_list.append(forecasts)

        return forecasts_list if len(series) > 1 else forecasts_list[0]

    def backtest(
        self,
        series: Union[TimeSeries, Sequence[TimeSeries]],
        past_covariates: Optional[Union[TimeSeries, Sequence[TimeSeries]]] = None,
        future_covariates: Optional[Union[TimeSeries, Sequence[TimeSeries]]] = None,
        historical_forecasts: Optional[Union[TimeSeries, Sequence[TimeSeries]]] = None,
        num_samples: int = 1,
        train_length: Optional[int] = None,
        start: Optional[Union[pd.Timestamp, float, int]] = None,
        forecast_horizon: int = 1,
        stride: int = 1,
        retrain: Union[bool, int, Callable[..., bool]] = True,
        overlap_end: bool = False,
        last_points_only: bool = False,
        metric: Union[
            Callable[[TimeSeries, TimeSeries], float],
            List[Callable[[TimeSeries, TimeSeries], float]],
        ] = metrics.mape,
        reduction: Union[Callable[[np.ndarray], float], None] = np.mean,
        verbose: bool = False,
        show_warnings: bool = True,
    ) -> Union[float, List[float], Sequence[float], List[Sequence[float]]]:
        """Compute error values that the model would have produced when
        used on (potentially multiple) `series`.

        If `historical_forecasts` are provided, the metric (given by the `metric` function) is evaluated directly on
        the forecast and the actual values. Otherwise, it repeatedly builds a training set: either expanding from the
        beginning  of `series` or moving with a fixed length `train_length`. It trains the current model on the
        training set, emits a forecast of length equal to `forecast_horizon`, and then moves the end of the training
        set forward by `stride` time steps. The metric is then evaluated on the forecast and the actual values.
        Finally, the method returns a `reduction` (the mean by default) of all these metric scores.

        By default, this method uses each historical forecast (whole) to compute error scores.
        If `last_points_only` is set to True, it will use only the last point of each historical
        forecast. In this case, no reduction is used.

        By default, this method always re-trains the models on the entire available history, corresponding to an
        expanding window strategy. If `retrain` is set to False (useful for models for which training might be
        time-consuming, such as deep learning models), the trained model will be used directly to emit the forecasts.

        Parameters
        ----------
        series
            The (or a sequence of) target time series used to successively train and evaluate the historical forecasts.
        past_covariates
            Optionally, one (or a sequence of) past-observed covariate series. This applies only if the model
            supports past covariates.
        future_covariates
            Optionally, one (or a sequence of) future-known covariate series. This applies only if the model
            supports future covariates.
        historical_forecasts
            Optionally, the (or a sequence of) historical forecasts time series to be evaluated. Corresponds to
            the output of :meth:`historical_forecasts() <ForecastingModel.historical_forecasts>`. If provided, will
            skip historical forecasting and ignore all parameters except `series`, `metric`, and `reduction`.
        num_samples
            Number of times a prediction is sampled from a probabilistic model. Use values `>1` only for probabilistic
            models.
        train_length
            Number of time steps in our training set (size of backtesting window to train on). Only effective when
            `retrain` is not ``False``. Default is set to `train_length=None` where it takes all available time steps
            up until prediction time, otherwise the moving window strategy is used. If larger than the number of time
            steps available, all steps up until prediction time are used, as in default case. Needs to be at least
            `min_train_series_length`.
        start
            Optionally, the first point in time at which a prediction is computed for a future time.
            This parameter supports: ``float``, ``int`` and ``pandas.Timestamp``, and ``None``.
            If a ``float``, the parameter will be treated as the proportion of the time series
            that should lie before the first prediction point.
            If an ``int``, the parameter will be treated as an integer index to the time index of
            `series` that will be used as first prediction time.
            If a ``pandas.Timestamp``, the time stamp will be used to determine the first prediction time
            directly.
            If ``None``, the first prediction time will automatically be set to:
                 - the first predictable point if `retrain` is ``False``, or `retrain` is a Callable and the first
                 predictable point is earlier than the first trainable point.

                 - the first trainable point if `retrain` is ``True`` or ``int`` (given `train_length`),
                 or `retrain` is a Callable and the first trainable point is earlier than the first predictable point.

                 - the first trainable point (given `train_length`) otherwise
            Note: Raises a ValueError if `start` yields a time outside the time index of `series`.
            Note: If `start` is outside the possible historical forecasting times, will ignore the parameter
            (default behavior with ``None``) and start at the first trainable/predictable point.
        forecast_horizon
            The forecast horizon for the point predictions.
        stride
            The number of time steps between two consecutive predictions.
        retrain
            Whether and/or on which condition to retrain the model before predicting.
            This parameter supports 3 different datatypes: ``bool``, (positive) ``int``, and
            ``Callable`` (returning a ``bool``).
            In the case of ``bool``: retrain the model at each step (`True`), or never retrains the model (`False`).
            In the case of ``int``: the model is retrained every `retrain` iterations.
            In the case of ``Callable``: the model is retrained whenever callable returns `True`.
            The callable must have the following positional arguments:

                - `counter` (int): current `retrain` iteration
                - `pred_time` (pd.Timestamp or int): timestamp of forecast time (end of the training series)
                - `train_series` (TimeSeries): train series up to `pred_time`
                - `past_covariates` (TimeSeries): past_covariates series up to `pred_time`
                - `future_covariates` (TimeSeries): future_covariates series up
                  to `min(pred_time + series.freq * forecast_horizon, series.end_time())`

            Note: if any optional `*_covariates` are not passed to `historical_forecast`, ``None`` will be passed
            to the corresponding retrain function argument.
            Note: some models do require being retrained every time and do not support anything other
            than `retrain=True`.
        overlap_end
            Whether the returned forecasts can go beyond the series' end or not.
        last_points_only
            Whether to use the whole historical forecasts or only the last point of each forecast to compute the error.
        metric
            A function or a list of function that takes two ``TimeSeries`` instances as inputs and returns an
            error value.
        reduction
            A function used to combine the individual error scores obtained when `last_points_only` is set to False.
            When providing several metric functions, the function will receive the argument `axis = 0` to obtain single
            value for each metric function.
            If explicitly set to `None`, the method will return a list of the individual error scores instead.
            Set to ``np.mean`` by default.
        verbose
            Whether to print progress.
        show_warnings
            Whether to show warnings related to parameters `start`, and `train_length`.

        Returns
        -------
        float or List[float] or List[List[float]]
            The (sequence of) error score on a series, or list of list containing error scores for each
            provided series and each sample.
        """
        if historical_forecasts is None:
            forecasts = self.historical_forecasts(
                series=series,
                past_covariates=past_covariates,
                future_covariates=future_covariates,
                num_samples=num_samples,
                train_length=train_length,
                start=start,
                forecast_horizon=forecast_horizon,
                stride=stride,
                retrain=retrain,
                overlap_end=overlap_end,
                last_points_only=last_points_only,
                verbose=verbose,
                show_warnings=show_warnings,
            )
        else:
            forecasts = historical_forecasts

        series = series2seq(series)
        if len(series) == 1:
            forecasts = [forecasts]

        if not isinstance(metric, list):
            metric = [metric]

        backtest_list = []
        for idx, target_ts in enumerate(series):
            if last_points_only:
                errors = [metric_f(target_ts, forecasts[idx]) for metric_f in metric]
                if len(errors) == 1:
                    errors = errors[0]
                backtest_list.append(errors)
            else:
                errors = [
                    [metric_f(target_ts, f) for metric_f in metric]
                    if len(metric) > 1
                    else metric[0](target_ts, f)
                    for f in forecasts[idx]
                ]

                if reduction is None:
                    backtest_list.append(errors)
                else:
                    backtest_list.append(reduction(np.array(errors), axis=0))

        return backtest_list if len(backtest_list) > 1 else backtest_list[0]

    @classmethod
    def gridsearch(
        model_class,
        parameters: dict,
        series: TimeSeries,
        past_covariates: Optional[TimeSeries] = None,
        future_covariates: Optional[TimeSeries] = None,
        forecast_horizon: Optional[int] = None,
        stride: int = 1,
        start: Union[pd.Timestamp, float, int] = 0.5,
        last_points_only: bool = False,
        show_warnings: bool = True,
        val_series: Optional[TimeSeries] = None,
        use_fitted_values: bool = False,
        metric: Callable[[TimeSeries, TimeSeries], float] = metrics.mape,
        reduction: Callable[[np.ndarray], float] = np.mean,
        verbose=False,
        n_jobs: int = 1,
        n_random_samples: Optional[Union[int, float]] = None,
    ) -> Tuple["ForecastingModel", Dict[str, Any], float]:
        """
        Find the best hyper-parameters among a given set using a grid search.

        This function has 3 modes of operation: Expanding window mode, split mode and fitted value mode.
        The three modes of operation evaluate every possible combination of hyper-parameter values
        provided in the `parameters` dictionary by instantiating the `model_class` subclass
        of ForecastingModel with each combination, and returning the best-performing model with regard
        to the `metric` function. The `metric` function is expected to return an error value,
        thus the model resulting in the smallest `metric` output will be chosen.

        The relationship of the training data and test data depends on the mode of operation.

        Expanding window mode (activated when `forecast_horizon` is passed):
        For every hyperparameter combination, the model is repeatedly trained and evaluated on different
        splits of `series`. This process is accomplished by using
        the :func:`backtest` function as a subroutine to produce historic forecasts starting from `start`
        that are compared against the ground truth values of `series`.
        Note that the model is retrained for every single prediction, thus this mode is slower.

        Split window mode (activated when `val_series` is passed):
        This mode will be used when the `val_series` argument is passed.
        For every hyper-parameter combination, the model is trained on `series` and
        evaluated on `val_series`.

        Fitted value mode (activated when `use_fitted_values` is set to `True`):
        For every hyper-parameter combination, the model is trained on `series`
        and evaluated on the resulting fitted values.
        Not all models have fitted values, and this method raises an error if the model doesn't have a `fitted_values`
        member. The fitted values are the result of the fit of the model on `series`. Comparing with the
        fitted values can be a quick way to assess the model, but one cannot see if the model is overfitting the series.

        Derived classes must ensure that a single instance of a model will not share parameters with the other
        instances, e.g., saving models in the same path. Otherwise, an unexpected behavior can arise while running
        several models in parallel (when ``n_jobs != 1``). If this cannot be avoided, then gridsearch
        should be redefined, forcing ``n_jobs = 1``.

        Currently this method only supports deterministic predictions (i.e. when models' predictions
        have only 1 sample).

        Parameters
        ----------
        model_class
            The ForecastingModel subclass to be tuned for 'series'.
        parameters
            A dictionary containing as keys hyperparameter names, and as values lists of values for the
            respective hyperparameter.
        series
            The target series used as input and target for training.
        past_covariates
            Optionally, a past-observed covariate series. This applies only if the model supports past covariates.
        future_covariates
            Optionally, a future-known covariate series. This applies only if the model supports future covariates.
        forecast_horizon
            The integer value of the forecasting horizon. Activates expanding window mode.
        stride
            The number of time steps between two consecutive predictions. Only used in expanding window mode.
        start
            The ``int``, ``float`` or ``pandas.Timestamp`` that represents the starting point in the time index
            of `series` from which predictions will be made to evaluate the model.
            For a detailed description of how the different data types are interpreted, please see the documentation
            for `ForecastingModel.backtest`. Only used in expanding window mode.
        last_points_only
            Whether to use the whole forecasts or only the last point of each forecast to compute the error. Only used
            in expanding window mode.
        show_warnings
            Whether to show warnings related to the `start` parameter. Only used in expanding window mode.
        val_series
            The TimeSeries instance used for validation in split mode. If provided, this series must start right after
            the end of `series`; so that a proper comparison of the forecast can be made.
        use_fitted_values
            If `True`, uses the comparison with the fitted values.
            Raises an error if ``fitted_values`` is not an attribute of `model_class`.
        metric
            A function that takes two TimeSeries instances as inputs (actual and prediction, in this order),
            and returns a float error value.
        reduction
            A reduction function (mapping array to float) describing how to aggregate the errors obtained
            on the different validation series when backtesting. By default it'll compute the mean of errors.
        verbose
            Whether to print progress.
        n_jobs
            The number of jobs to run in parallel. Parallel jobs are created only when there are two or more parameters
            combinations to evaluate. Each job will instantiate, train, and evaluate a different instance of the model.
            Defaults to `1` (sequential). Setting the parameter to `-1` means using all the available cores.
        n_random_samples
            The number/ratio of hyperparameter combinations to select from the full parameter grid. This will perform
            a random search instead of using the full grid.
            If an integer, `n_random_samples` is the number of parameter combinations selected from the full grid and
            must be between `0` and the total number of parameter combinations.
            If a float, `n_random_samples` is the ratio of parameter combinations selected from the full grid and must
            be between `0` and `1`. Defaults to `None`, for which random selection will be ignored.

        Returns
        -------
        ForecastingModel, Dict, float
            A tuple containing an untrained `model_class` instance created from the best-performing hyper-parameters,
            along with a dictionary containing these best hyper-parameters,
            and metric score for the best hyper-parameters.
        """
        raise_if_not(
            (forecast_horizon is not None)
            + (val_series is not None)
            + use_fitted_values
            == 1,
            "Please pass exactly one of the arguments 'forecast_horizon', "
            "'val_target_series' or 'use_fitted_values'.",
            logger,
        )

        if use_fitted_values:
            raise_if_not(
                hasattr(model_class(), "fitted_values"),
                "The model must have a fitted_values attribute to compare with the train TimeSeries",
                logger,
            )

        elif val_series is not None:
            raise_if_not(
                series.width == val_series.width,
                "Training and validation series require the same number of components.",
                logger,
            )

        # TODO: here too I'd say we can leave these checks to the models
        # if covariates is not None:
        #     raise_if_not(series.has_same_time_as(covariates), 'The provided series and covariates must have the '
        #                                                       'same time axes.')

        # compute all hyperparameter combinations from selection
        params_cross_product = list(product(*parameters.values()))

        # If n_random_samples has been set, randomly select a subset of the full parameter cross product to search with
        if n_random_samples is not None:
            params_cross_product = model_class._sample_params(
                params_cross_product, n_random_samples
            )

        # iterate through all combinations of the provided parameters and choose the best one
        iterator = _build_tqdm_iterator(
            zip(params_cross_product), verbose, total=len(params_cross_product)
        )

        def _evaluate_combination(param_combination) -> float:
            param_combination_dict = dict(
                list(zip(parameters.keys(), param_combination))
            )
            if param_combination_dict.get("model_name", None):
                current_time = time.strftime("%Y-%m-%d_%H.%M.%S.%f", time.localtime())
                param_combination_dict[
                    "model_name"
                ] = f"{current_time}_{param_combination_dict['model_name']}"

            model = model_class(**param_combination_dict)
            if use_fitted_values:  # fitted value mode
                model._fit_wrapper(series, past_covariates, future_covariates)
                fitted_values = TimeSeries.from_times_and_values(
                    series.time_index, model.fitted_values
                )
                error = metric(series, fitted_values)
            elif val_series is None:  # expanding window mode
                error = model.backtest(
                    series=series,
                    past_covariates=past_covariates,
                    future_covariates=future_covariates,
                    num_samples=1,
                    start=start,
                    forecast_horizon=forecast_horizon,
                    stride=stride,
                    metric=metric,
                    reduction=reduction,
                    last_points_only=last_points_only,
                    verbose=verbose,
                    show_warnings=show_warnings,
                )
            else:  # split mode
                model._fit_wrapper(series, past_covariates, future_covariates)
                pred = model._predict_wrapper(
                    len(val_series),
                    series,
                    past_covariates,
                    future_covariates,
                    num_samples=1,
                    verbose=verbose,
                )
                error = metric(val_series, pred)

            return float(error)

        errors: List[float] = _parallel_apply(
            iterator, _evaluate_combination, n_jobs, {}, {}
        )

        min_error = min(errors)

        best_param_combination = dict(
            list(zip(parameters.keys(), params_cross_product[errors.index(min_error)]))
        )

        logger.info("Chosen parameters: " + str(best_param_combination))

        return model_class(**best_param_combination), best_param_combination, min_error

    def residuals(
        self,
        series: Union[TimeSeries, Sequence[TimeSeries]],
        past_covariates: Optional[Union[TimeSeries, Sequence[TimeSeries]]] = None,
        future_covariates: Optional[Union[TimeSeries, Sequence[TimeSeries]]] = None,
        forecast_horizon: int = 1,
        retrain: bool = True,
        verbose: bool = False,
    ) -> Union[TimeSeries, Sequence[TimeSeries]]:
        """Compute the residuals produced by this model on a (or sequence of) univariate  time series.

        This function computes the difference between the actual observations from `series` and the fitted values
        vector `p` obtained by training the model on `series`. For every index `i` in `series`, `p[i]` is computed
        by training the model on ``series[:(i - forecast_horizon)]`` and forecasting `forecast_horizon` into the future.
        (`p[i]` will be set to the last value of the predicted series.)
        The vector of residuals will be shorter than `series` due to the minimum training series length required by the
        model and the gap introduced by `forecast_horizon`. Most commonly, the term "residuals" implies a value for
        `forecast_horizon` of 1; but this can be configured.

        This method works only on univariate series. It uses the median
        prediction (when dealing with stochastic forecasts).

        Parameters
        ----------
        series
            The univariate TimeSeries instance which the residuals will be computed for.
        past_covariates
            One or several past-observed covariate time series.
        future_covariates
            One or several future-known covariate time series.
        forecast_horizon
            The forecasting horizon used to predict each fitted value.
        retrain
            Whether to train the model at each iteration, for models that support it.
            If False, the model is not trained at all. Default: True
        verbose
            Whether to print progress.

        Returns
        -------
        TimeSeries (or Sequence[TimeSeries])
            The vector of residuals.
        """

        series = series2seq(series)
        past_covariates = series2seq(past_covariates)
        future_covariates = series2seq(future_covariates)

        raise_if_not(
            all([serie.is_univariate for serie in series]),
            "Each series in the sequence must be univariate.",
            logger,
        )

        residuals_list = []
        # compute residuals
        for idx, target_ts in enumerate(series):
            # get first index not contained in the first training set
            first_index = target_ts.time_index[self.min_train_series_length]

            # compute fitted values
            forecasts = self.historical_forecasts(
                series=target_ts,
                past_covariates=past_covariates[idx] if past_covariates else None,
                future_covariates=future_covariates[idx] if future_covariates else None,
                start=first_index,
                forecast_horizon=forecast_horizon,
                stride=1,
                retrain=retrain,
                last_points_only=True,
                verbose=verbose,
            )
            series_trimmed = target_ts.slice_intersect(forecasts)
            residuals_list.append(
                series_trimmed
                - (
                    forecasts.quantile_timeseries(quantile=0.5)
                    if forecasts.is_stochastic
                    else forecasts
                )
            )

        return residuals_list if len(residuals_list) > 1 else residuals_list[0]

    def initialize_encoders(self) -> SequentialEncoder:
        """instantiates the SequentialEncoder object based on self._model_encoder_settings and parameter
        ``add_encoders`` used at model creation"""
        (
            input_chunk_length,
            output_chunk_length,
            takes_past_covariates,
            takes_future_covariates,
            lags_past_covariates,
            lags_future_covariates,
        ) = self._model_encoder_settings

        return SequentialEncoder(
            add_encoders=self.add_encoders,
            input_chunk_length=input_chunk_length,
            output_chunk_length=output_chunk_length,
            lags_past_covariates=lags_past_covariates,
            lags_future_covariates=lags_future_covariates,
            takes_past_covariates=takes_past_covariates,
            takes_future_covariates=takes_future_covariates,
        )

    def generate_fit_encodings(
        self,
        series: Union[TimeSeries, Sequence[TimeSeries]],
        past_covariates: Optional[Union[TimeSeries, Sequence[TimeSeries]]] = None,
        future_covariates: Optional[Union[TimeSeries, Sequence[TimeSeries]]] = None,
    ) -> Tuple[
        Union[TimeSeries, Sequence[TimeSeries]], Union[TimeSeries, Sequence[TimeSeries]]
    ]:
        """Generates the covariate encodings that were used/generated for fitting the model and returns a tuple of
        past, and future covariates series with the original and encoded covariates stacked together. The encodings are
        generated by the encoders defined at model creation with parameter `add_encoders`. Pass the same `series`,
        `past_covariates`, and  `future_covariates` that you used to train/fit the model.

        Parameters
        ----------
        series
            The series or sequence of series with the target values used when fitting the model.
        past_covariates
            Optionally, the series or sequence of series with the past-observed covariates used when fitting the model.
        future_covariates
            Optionally, the series or sequence of series with the future-known covariates used when fitting the model.

        Returns
        -------
        Tuple[Union[TimeSeries, Sequence[TimeSeries]], Union[TimeSeries, Sequence[TimeSeries]]]
            A tuple of (past covariates, future covariates). Each covariate contains the original as well as the
            encoded covariates.
        """
        raise_if(
            self.encoders is None or not self.encoders.encoding_available,
            "Encodings are not available. Consider adding parameter `add_encoders` at model creation and fitting the "
            "model with `model.fit()` before.",
            logger=logger,
        )
        return self.encoders.encode_train(
            target=series,
            past_covariates=past_covariates,
            future_covariates=future_covariates,
        )

    def generate_predict_encodings(
        self,
        n: int,
        series: Union[TimeSeries, Sequence[TimeSeries]],
        past_covariates: Optional[Union[TimeSeries, Sequence[TimeSeries]]] = None,
        future_covariates: Optional[Union[TimeSeries, Sequence[TimeSeries]]] = None,
    ) -> Tuple[
        Union[TimeSeries, Sequence[TimeSeries]], Union[TimeSeries, Sequence[TimeSeries]]
    ]:
        """Generates covariate encodings for the inference/prediction set and returns a tuple of past, and future
        covariates series with the original and encoded covariates stacked together. The encodings are generated by the
        encoders defined at model creation with parameter `add_encoders`. Pass the same `series`, `past_covariates`,
        and `future_covariates` that you intend to use for prediction.

        Parameters
        ----------
        n
            The number of prediction time steps after the end of `series` intended to be used for prediction.
        series
            The series or sequence of series with target values intended to be used for prediction.
        past_covariates
            Optionally, the past-observed covariates series intended to be used for prediction. The dimensions must
            match those of the covariates used for training.
        future_covariates
            Optionally, the future-known covariates series intended to be used for prediction. The dimensions must
            match those of the covariates used for training.

        Returns
        -------
        Tuple[Union[TimeSeries, Sequence[TimeSeries]], Union[TimeSeries, Sequence[TimeSeries]]]
            A tuple of (past covariates, future covariates). Each covariate contains the original as well as the
            encoded covariates.
        """
        raise_if(
            self.encoders is None or not self.encoders.encoding_available,
            "Encodings are not available. Consider adding parameter `add_encoders` at model creation and fitting the "
            "model with `model.fit()` before.",
            logger=logger,
        )
        return self.encoders.encode_inference(
            n=n,
            target=series,
            past_covariates=past_covariates,
            future_covariates=future_covariates,
        )

    @property
    @abstractmethod
    def _model_encoder_settings(
        self,
    ) -> Tuple[
        Optional[int],
        Optional[int],
        bool,
        bool,
        Optional[List[int]],
        Optional[List[int]],
    ]:
        """Abstract property that returns model specific encoder settings that are used to initialize the encoders.

        Must return Tuple (input_chunk_length, output_chunk_length, takes_past_covariates, takes_future_covariates,
        lags_past_covariates, lags_future_covariates).
        """
        pass

    @classmethod
    def _sample_params(model_class, params, n_random_samples):
        """Select the absolute number of samples randomly if an integer has been supplied. If a float has been
        supplied, select a fraction"""

        if isinstance(n_random_samples, int):
            raise_if_not(
                (n_random_samples > 0) and (n_random_samples <= len(params)),
                "If supplied as an integer, n_random_samples must be greater than 0 and less"
                "than or equal to the size of the cartesian product of the hyperparameters.",
            )
            return sample(params, n_random_samples)

        if isinstance(n_random_samples, float):
            raise_if_not(
                (n_random_samples > 0.0) and (n_random_samples <= 1.0),
                "If supplied as a float, n_random_samples must be greater than 0.0 and less than 1.0.",
            )
            return sample(params, int(n_random_samples * len(params)))

    def _extract_model_creation_params(self):
        """extracts immutable model creation parameters from `ModelMeta` and deletes reference."""
        model_params = copy.deepcopy(self._model_call)
        del self.__class__._model_call
        return model_params

    def untrained_model(self):
        return self.__class__(**copy.deepcopy(self.model_params))

    @property
    def model_params(self) -> dict:
        return (
            self._model_params if hasattr(self, "_model_params") else self._model_call
        )

    @classmethod
    def _default_save_path(cls) -> str:
        return f"{cls.__name__}_{datetime.datetime.now().strftime('%Y-%m-%d_%H_%M_%S')}"

    def save(self, path: Optional[Union[str, BinaryIO]] = None, **pkl_kwargs) -> None:
        """
        Saves the model under a given path or file handle.

        Example for saving and loading a :class:`RegressionModel`:

            .. highlight:: python
            .. code-block:: python

                from darts.models import RegressionModel

                model = RegressionModel(lags=4)

                model.save("my_model.pkl")
                model_loaded = RegressionModel.load("my_model.pkl")
            ..

        Parameters
        ----------
        path
            Path or file handle under which to save the model at its current state. If no path is specified, the model
            is automatically saved under ``"{ModelClass}_{YYYY-mm-dd_HH_MM_SS}.pkl"``.
            E.g., ``"RegressionModel_2020-01-01_12_00_00.pkl"``.
        pkl_kwargs
            Keyword arguments passed to `pickle.dump()`
        """

        if path is None:
            # default path
            path = self._default_save_path() + ".pkl"

        if isinstance(path, str):
            # save the whole object using pickle
            with open(path, "wb") as handle:
                pickle.dump(obj=self, file=handle, **pkl_kwargs)
        else:
            # save the whole object using pickle
            pickle.dump(obj=self, file=path, **pkl_kwargs)

    @staticmethod
    def load(path: Union[str, BinaryIO]) -> "ForecastingModel":
        """
        Loads the model from a given path or file handle.

        Parameters
        ----------
        path
            Path or file handle from which to load the model.
        """

        if isinstance(path, str):
            raise_if_not(
                os.path.exists(path),
                f"The file {path} doesn't exist",
                logger,
            )

            with open(path, "rb") as handle:
                model = pickle.load(file=handle)
        else:
            model = pickle.load(file=path)

        return model

    def _assert_univariate(self, series: TimeSeries):
        if not series.is_univariate:
            raise_log(
                ValueError(
                    f"Model `{self.__class__.__name__}` only supports univariate TimeSeries instances"
                ),
                logger=logger,
            )

    def _assert_multivariate(self, series: TimeSeries):
        if series.is_univariate:
            raise_log(
                ValueError(
                    f"Model `{self.__class__.__name__}` only supports multivariate TimeSeries instances"
                ),
                logger=logger,
            )

    def __repr__(self):
        """
        Get full description for this estimator (includes all params).
        """
        return self._get_model_description_string(True)

    def __str__(self):
        """
        Get short description for this estimator (only includes params with non-default values).
        """
        return self._get_model_description_string(False)

    def _get_model_description_string(self, include_default_params):
        """
        Get model description string of structure `model_name`(`model_param_key_value_pairs`).

        Parameters
        ----------
        include_default_params : bool,
            If True, will include params with default values in the description.

        Returns
        -------
        description : String
            Model description.
        """
        default_model_params = self._get_default_model_params()
        changed_model_params = [
            (k, v)
            for k, v in self.model_params.items()
            if include_default_params or v != default_model_params.get(k, None)
        ]

        model_name = self.__class__.__name__
        params_string = ", ".join([f"{k}={str(v)}" for k, v in changed_model_params])
        return f"{model_name}({params_string})"

    @classmethod
    def _get_default_model_params(cls):
        """Get parameter key : default_value pairs for the estimator"""
        init_signature = inspect.signature(cls.__init__)
        # Consider the constructor parameters excluding 'self'
        return {
            p.name: p.default
            for p in init_signature.parameters.values()
            if p.name != "self"
        }


class LocalForecastingModel(ForecastingModel, ABC):
    """The base class for "local" forecasting models, handling only single univariate time series.

    Local Forecasting Models (LFM) are models that can be trained on a single univariate target series only. In Darts,
    most models in this category tend to be simpler statistical models (such as ETS or FFT). LFMs usually train on
    the entire target series supplied when calling :func:`fit()` at once. They can also predict in one go with
    :func:`predict()` for any number of predictions `n` after the end of the training series.

    All implementations must implement the `fit()` and `predict()` methods.
    """

    def __init__(self, add_encoders: Optional[dict] = None):
        super().__init__(add_encoders=add_encoders)

    @property
    def _model_encoder_settings(
        self,
    ) -> Tuple[
        Optional[int],
        Optional[int],
        bool,
        bool,
        Optional[List[int]],
        Optional[List[int]],
    ]:
        return None, None, False, False, None, None

    @abstractmethod
    def fit(self, series: TimeSeries) -> "LocalForecastingModel":
        super().fit(series)
        series._assert_deterministic()

    @property
    def extreme_lags(
        self,
    ) -> Tuple[
        Optional[int],
        Optional[int],
        Optional[int],
        Optional[int],
        Optional[int],
        Optional[int],
    ]:
        # TODO: LocalForecastingModels do not yet handle extreme lags properly. Especially
        #  TransferableFutureCovariatesLocalForecastingModel, where there is a difference between fit and predict mode)
        #  do not yet. In general, Local models train on the entire series (input=output), different to Global models
        #  that use an input to predict an output.
        return -self.min_train_series_length, -1, None, None, None, None


class GlobalForecastingModel(ForecastingModel, ABC):
    """The base class for "global" forecasting models, handling several time series and optional covariates.

    Global forecasting models expand upon the functionality of `ForecastingModel` in 4 ways:
    1. Models can be fitted on many series (multivariate or univariate) with different indices.
    2. The input series used by :func:`predict()` can be different from the series used to fit the model.
    3. Covariates can be supported (multivariate or univariate).
    4. They can allow for multivariate target series and covariates.

    The name "global" stems from the fact that a training set of a forecasting model of this class is not constrained
    to a temporally contiguous, "local", time series.

    All implementations must implement the :func:`fit()` and :func:`predict()` methods.
    The :func:`fit()` method is meant to train the model on one or several training time series, along with optional
    covariates.

    If :func:`fit()` has been called with only one training and covariate series as argument, then
    calling :func:`predict()` will forecast the future of this series. Otherwise, the user has to
    provide to :func:`predict()` the series they want to forecast, as well as covariates, if needed.
    """

    def __init__(self, add_encoders: Optional[dict] = None):
        super().__init__(add_encoders=add_encoders)

    @abstractmethod
    def fit(
        self,
        series: Union[TimeSeries, Sequence[TimeSeries]],
        past_covariates: Optional[Union[TimeSeries, Sequence[TimeSeries]]] = None,
        future_covariates: Optional[Union[TimeSeries, Sequence[TimeSeries]]] = None,
    ) -> "GlobalForecastingModel":
        """Fit/train the model on (potentially multiple) series.

        Optionally, one or multiple past and/or future covariates series can be provided as well.
        The number of covariates series must match the number of target series.

        Parameters
        ----------
        series
            One or several target time series. The model will be trained to forecast these time series.
            The series may or may not be multivariate, but if multiple series are provided
            they must have the same number of components.
        past_covariates
            One or several past-observed covariate time series. These time series will not be forecast, but can
            be used by some models as an input. The covariate(s) may or may not be multivariate, but if multiple
            covariates are provided they must have the same number of components. If `past_covariates` is provided,
            it must contain the same number of series as `series`.
        future_covariates
            One or several future-known covariate time series. These time series will not be forecast, but can
            be used by some models as an input. The covariate(s) may or may not be multivariate, but if multiple
            covariates are provided they must have the same number of components. If `future_covariates` is provided,
            it must contain the same number of series as `series`.

        Returns
        -------
        self
            Fitted model.
        """

        if isinstance(series, TimeSeries):
            # if only one series is provided, save it for prediction time (including covariates, if available)
            self.training_series = series
            if past_covariates is not None:
                self.past_covariate_series = past_covariates
            if future_covariates is not None:
                self.future_covariate_series = future_covariates
            if (
                series.static_covariates is not None
                and self.supports_static_covariates
                and self.considers_static_covariates
            ):
                self.static_covariates = series.static_covariates
        else:
            if past_covariates is not None:
                self._expect_past_covariates = True
            if future_covariates is not None:
                self._expect_future_covariates = True
            if (
                get_single_series(series).static_covariates is not None
                and self.supports_static_covariates
                and self.considers_static_covariates
            ):
                self.static_covariates = series[0].static_covariates
                self._expect_static_covariates = True

        if past_covariates is not None:
            self._uses_past_covariates = True
        if future_covariates is not None:
            self._uses_future_covariates = True
        if (
            get_single_series(series).static_covariates is not None
            and self.supports_static_covariates
            and self.considers_static_covariates
        ):
            self._uses_static_covariates = True
        self._fit_called = True

    @abstractmethod
    def predict(
        self,
        n: int,
        series: Optional[Union[TimeSeries, Sequence[TimeSeries]]] = None,
        past_covariates: Optional[Union[TimeSeries, Sequence[TimeSeries]]] = None,
        future_covariates: Optional[Union[TimeSeries, Sequence[TimeSeries]]] = None,
        num_samples: int = 1,
        verbose: bool = False,
    ) -> Union[TimeSeries, Sequence[TimeSeries]]:
        """Forecasts values for `n` time steps after the end of the series.

        If :func:`fit()` has been called with only one ``TimeSeries`` as argument, then the `series` argument of
        this function is optional, and it will simply produce the next `horizon` time steps forecast.
        The `past_covariates` and `future_covariates` arguments also don't have to be provided again in this case.

        If :func:`fit()` has been called with `series` specified as a ``Sequence[TimeSeries]`` (i.e., the model
        has been trained on multiple time series), the `series` argument must be specified.

        When the `series` argument is specified, this function will compute the next `n` time steps forecasts
        for the simple series (or for each series in the sequence) given by `series`.

        If multiple past or future covariates were specified during the training, some corresponding covariates must
        also be specified here. For every input in `series` a matching (past and/or future) covariate time series
        has to be provided.

        Parameters
        ----------
        n
            Forecast horizon - the number of time steps after the end of the series for which to produce predictions.
        series
            The series whose future values will be predicted.
        past_covariates
            One past-observed covariate time series for every input time series in `series`. They must match the
            past covariates that have been used with the :func:`fit()` function for training in terms of dimension.
        future_covariates
            One future-known covariate time series for every input time series in `series`. They must match the
            past covariates that have been used with the :func:`fit()` function for training in terms of dimension.
        num_samples
            Number of times a prediction is sampled from a probabilistic model. Should be left set to 1
            for deterministic models.

        Returns
        -------
        Union[TimeSeries, Sequence[TimeSeries]]
            If `series` is not specified, this function returns a single time series containing the `n`
            next points after then end of the training series.
            If `series` is given and is a simple ``TimeSeries``, this function returns the `n` next points
            after the end of `series`.
            If `series` is given and is a sequence of several time series, this function returns
            a sequence where each element contains the corresponding `n` points forecasts.
        """
        super().predict(n, num_samples)
        if self.uses_past_covariates and past_covariates is None:
            raise_log(
                ValueError(
                    "The model has been trained with past covariates. Some matching past_covariates "
                    "have to be provided to `predict()`."
                )
            )
        if self.uses_future_covariates and future_covariates is None:
            raise_log(
                ValueError(
                    "The model has been trained with future covariates. Some matching future_covariates "
                    "have to be provided to `predict()`."
                )
            )
        if (
            self.uses_static_covariates
            and get_single_series(series).static_covariates is None
        ):
            raise_log(
                ValueError(
                    "The model has been trained with static covariates. Some matching static covariates "
                    "must be embedded in the target `series` passed to `predict()`."
                )
            )

    def _predict_wrapper(
        self,
        n: int,
        series: TimeSeries,
        past_covariates: Optional[TimeSeries],
        future_covariates: Optional[TimeSeries],
        num_samples: int,
        verbose: bool = False,
    ) -> TimeSeries:
        return self.predict(
            n,
            series,
            past_covariates=past_covariates,
            future_covariates=future_covariates,
            num_samples=num_samples,
            verbose=verbose,
        )

    def _fit_wrapper(
        self,
        series: TimeSeries,
        past_covariates: Optional[TimeSeries],
        future_covariates: Optional[TimeSeries],
    ):
        self.fit(
            series=series,
            past_covariates=past_covariates if self.supports_past_covariates else None,
            future_covariates=future_covariates
            if self.supports_future_covariates
            else None,
        )

    def _supports_non_retrainable_historical_forecasts(self) -> bool:
        """GlobalForecastingModel supports historical forecasts without retraining the model"""
        return True


class FutureCovariatesLocalForecastingModel(LocalForecastingModel, ABC):
    """The base class for future covariates "local" forecasting models, handling single uni- or multivariate target
    and optional future covariates time series.

    Future Covariates Local Forecasting Models (FC-LFM) are models that can be trained on a single uni- or multivariate
    target and optional future covariates series. In Darts, most models in this category tend to be simpler statistical
    models (such as ARIMA). FC-LFMs usually train on the entire target and future covariates series supplied when
    calling :func:`fit()` at once. They can also predict in one go with :func:`predict()` for any number of predictions
    `n` after the end of the training series. When using future covariates, the values for the future `n` prediction
    steps must be given in the covariate series.

    All implementations must implement the :func:`_fit()` and :func:`_predict()` methods.
    """

    def fit(self, series: TimeSeries, future_covariates: Optional[TimeSeries] = None):
        """Fit/train the model on the (single) provided series.

        Optionally, a future covariates series can be provided as well.

        Parameters
        ----------
        series
            The model will be trained to forecast this time series. Can be multivariate if the model supports it.
        future_covariates
            A time series of future-known covariates. This time series will not be forecasted, but can be used by
            some models as an input. It must contain at least the same time steps/indices as the target `series`.
            If it is longer than necessary, it will be automatically trimmed.

        Returns
        -------
        self
            Fitted model.
        """

        if future_covariates is not None:
            if not series.has_same_time_as(future_covariates):
                # fit() expects future_covariates to have same time as the target, so we intersect it here
                future_covariates = future_covariates.slice_intersect(series)

            raise_if_not(
                series.has_same_time_as(future_covariates),
                "The provided `future_covariates` series must contain at least the same time steps/"
                "indices as the target `series`.",
                logger=logger,
            )
            self._expect_future_covariates = True

        self.encoders = self.initialize_encoders()
        if self.encoders.encoding_available:
            _, future_covariates = self.generate_fit_encodings(
                series=series,
                past_covariates=None,
                future_covariates=future_covariates,
            )

        super().fit(series)

        return self._fit(series, future_covariates=future_covariates)

    @abstractmethod
    def _fit(self, series: TimeSeries, future_covariates: Optional[TimeSeries] = None):
        """Fits/trains the model on the provided series.
        DualCovariatesModels must implement the fit logic in this method.
        """
        pass

    def predict(
        self,
        n: int,
        future_covariates: Optional[TimeSeries] = None,
        num_samples: int = 1,
        **kwargs,
    ) -> TimeSeries:
        """Forecasts values for `n` time steps after the end of the training series.

        If some future covariates were specified during the training, they must also be specified here.

        Parameters
        ----------
        n
            Forecast horizon - the number of time steps after the end of the series for which to produce predictions.
        future_covariates
            The time series of future-known covariates which can be fed as input to the model. It must correspond to
            the covariate time series that has been used with the :func:`fit()` method for training, and it must
            contain at least the next `n` time steps/indices after the end of the training target series.
        num_samples
            Number of times a prediction is sampled from a probabilistic model. Should be left set to 1
            for deterministic models.

        Returns
        -------
        TimeSeries, a single time series containing the `n` next points after then end of the training series.
        """

        super().predict(n, num_samples)

        # avoid generating encodings again if subclass has already generated them
        if not self._supress_generate_predict_encoding:
            self._verify_passed_predict_covariates(future_covariates)
            if self.encoders is not None and self.encoders.encoding_available:
                _, future_covariates = self.generate_predict_encodings(
                    n=n,
                    series=self.training_series,
                    past_covariates=None,
                    future_covariates=future_covariates,
                )

        if future_covariates is not None:
            start = self.training_series.end_time() + self.training_series.freq

            invalid_time_span_error = (
                f"For the given forecasting horizon `n={n}`, the provided `future_covariates` "
                f"series must contain at least the next `n={n}` time steps/indices after the "
                f"end of the target `series` that was used to train the model."
            )

            # we raise an error here already to avoid getting error from empty TimeSeries creation
            raise_if_not(
                future_covariates.end_time() >= start, invalid_time_span_error, logger
            )

            offset = (
                n - 1
                if isinstance(future_covariates.time_index, pd.DatetimeIndex)
                else n
            )
            future_covariates = future_covariates.slice(
                start, start + offset * self.training_series.freq
            )

            raise_if_not(
                len(future_covariates) == n,
                invalid_time_span_error,
                logger,
            )

        return self._predict(
            n, future_covariates=future_covariates, num_samples=num_samples, **kwargs
        )

    @abstractmethod
    def _predict(
        self,
        n: int,
        future_covariates: Optional[TimeSeries] = None,
        num_samples: int = 1,
        verbose: bool = False,
        **kwargs,
    ) -> TimeSeries:
        """Forecasts values for a certain number of time steps after the end of the series.
        DualCovariatesModels must implement the predict logic in this method.
        """
        pass

    def _fit_wrapper(
        self,
        series: TimeSeries,
        past_covariates: Optional[TimeSeries],
        future_covariates: Optional[TimeSeries],
    ):
        self.fit(series, future_covariates=future_covariates)

    def _predict_wrapper(
        self,
        n: int,
        series: TimeSeries,
        past_covariates: Optional[TimeSeries],
        future_covariates: Optional[TimeSeries],
        num_samples: int,
        verbose: bool = False,
    ) -> TimeSeries:
        return self.predict(
            n,
            future_covariates=future_covariates,
            num_samples=num_samples,
            verbose=verbose,
        )

    @property
    def _model_encoder_settings(
        self,
    ) -> Tuple[
        Optional[int],
        Optional[int],
        bool,
        bool,
        Optional[List[int]],
        Optional[List[int]],
    ]:
        return None, None, False, True, None, None

    def _verify_passed_predict_covariates(self, future_covariates):
        """Simple check if user supplied/did not supply covariates as done at fitting time."""
        if self._expect_future_covariates and future_covariates is None:
            raise_log(
                ValueError(
                    "The model has been trained with `future_covariates` variable. Some matching "
                    "`future_covariates` variables have to be provided to `predict()`."
                )
            )
        if not self._expect_future_covariates and future_covariates is not None:
            raise_log(
                ValueError(
                    "The model has been trained without `future_covariates` variable, but the "
                    "`future_covariates` parameter provided to `predict()` is not None.",
                )
            )

    @property
    def _supress_generate_predict_encoding(self) -> bool:
        """Controls wether encodings should be generated in :func:`FutureCovariatesLocalForecastingModel.predict()``"""
        return False

    @property
    def extreme_lags(
        self,
    ) -> Tuple[
        Optional[int],
        Optional[int],
        Optional[int],
        Optional[int],
        Optional[int],
        Optional[int],
    ]:
        # TODO: LocalForecastingModels do not yet handle extreme lags properly. Especially
        #  TransferableFutureCovariatesLocalForecastingModel, where there is a difference between fit and predict mode)
        #  do not yet. In general, Local models train on the entire series (input=output), different to Global models
        #  that use an input to predict an output.
        return -self.min_train_series_length, -1, None, None, 0, 0


class TransferableFutureCovariatesLocalForecastingModel(
    FutureCovariatesLocalForecastingModel, ABC
):
    """The base class for transferable future covariates "local" forecasting models, handling single uni- or
    multivariate target and optional future covariates time series. Additionally, at prediction time, it can be
    applied to new data unrelated to the original series used for fitting the model.

    Transferable Future Covariates Local Forecasting Models (TFC-LFM) are models that can be trained on a single uni-
    or multivariate target and optional future covariates series. Additionally, at prediction time, it can be applied
    to new data unrelated to the original series used for fitting the model. Currently in Darts, all models in this
    category wrap to statsmodel models such as VARIMA. TFC-LFMs usually train on the entire target and future covariates
    series supplied when calling :func:`fit()` at once. They can also predict in one go with :func:`predict()`
    for any number of predictions `n` after the end of the training series. When using future covariates, the values
    for the future `n` prediction steps must be given in the covariate series.

    All implementations must implement the :func:`_fit()` and :func:`_predict()` methods.
    """

    def predict(
        self,
        n: int,
        series: Optional[TimeSeries] = None,
        future_covariates: Optional[TimeSeries] = None,
        num_samples: int = 1,
        **kwargs,
    ) -> TimeSeries:
        """If the `series` parameter is not set, forecasts values for `n` time steps after the end of the training
        series. If some future covariates were specified during the training, they must also be specified here.

        If the `series` parameter is set, forecasts values for `n` time steps after the end of the new target
        series. If some future covariates were specified during the training, they must also be specified here.

        Parameters
        ----------
        n
            Forecast horizon - the number of time steps after the end of the series for which to produce predictions.
        series
            Optionally, a new target series whose future values will be predicted. Defaults to `None`, meaning that the
            model will forecast the future value of the training series.
        future_covariates
            The time series of future-known covariates which can be fed as input to the model. It must correspond to
            the covariate time series that has been used with the :func:`fit()` method for training.

            If `series` is not set, it must contain at least the next `n` time steps/indices after the end of the
            training target series. If `series` is set, it must contain at least the time steps/indices corresponding
            to the new target series (historic future covariates), plus the next `n` time steps/indices after the end.
        num_samples
            Number of times a prediction is sampled from a probabilistic model. Should be left set to 1
            for deterministic models.

        Returns
        -------
        TimeSeries, a single time series containing the `n` next points after then end of the training series.
        """
        self._verify_passed_predict_covariates(future_covariates)
        if self.encoders is not None and self.encoders.encoding_available:
            _, future_covariates = self.generate_predict_encodings(
                n=n,
                series=series if series is not None else self.training_series,
                past_covariates=None,
                future_covariates=future_covariates,
            )

        historic_future_covariates = None

        if series is not None and future_covariates:
            raise_if_not(
                future_covariates.start_time() <= series.start_time()
                and future_covariates.end_time() >= series.end_time() + n * series.freq,
                "The provided `future_covariates` related to the new target series must contain at least the same time"
                "steps/indices as the target `series` + `n`.",
                logger,
            )
            # splitting the future covariates
            (
                historic_future_covariates,
                future_covariates,
            ) = future_covariates.split_after(series.end_time())

            # in case future covariates have more values on the left end side that we don't need
            if not series.has_same_time_as(historic_future_covariates):
                historic_future_covariates = historic_future_covariates.slice_intersect(
                    series
                )

        # FutureCovariatesLocalForecastingModel performs some checks on self.training_series. We temporary replace
        # that with the new ts
        if series is not None:
            self._orig_training_series = self.training_series
            self.training_series = series

        result = super().predict(
            n=n,
            series=series,
            historic_future_covariates=historic_future_covariates,
            future_covariates=future_covariates,
            num_samples=num_samples,
            **kwargs,
        )

        # restoring the original training ts
        if series is not None:
            self.training_series = self._orig_training_series

        return result

    def generate_predict_encodings(
        self,
        n: int,
        series: Union[TimeSeries, Sequence[TimeSeries]],
        past_covariates: Optional[Union[TimeSeries, Sequence[TimeSeries]]] = None,
        future_covariates: Optional[Union[TimeSeries, Sequence[TimeSeries]]] = None,
    ) -> Tuple[
        Union[TimeSeries, Sequence[TimeSeries]], Union[TimeSeries, Sequence[TimeSeries]]
    ]:
        raise_if(
            self.encoders is None or not self.encoders.encoding_available,
            "Encodings are not available. Consider adding parameter `add_encoders` at model creation and fitting the "
            "model with `model.fit()` before.",
            logger=logger,
        )
        _, future_covariates_future = self.encoders.encode_inference(
            n=n,
            target=series,
            past_covariates=past_covariates,
            future_covariates=future_covariates,
        )

        if future_covariates is not None:
            return _, future_covariates_future

        _, future_covariates_historic = self.encoders.encode_train(
            target=series,
            past_covariates=past_covariates,
            future_covariates=future_covariates,
        )
        return _, future_covariates_historic.append(future_covariates_future)

    @abstractmethod
    def _predict(
        self,
        n: int,
        series: Optional[TimeSeries] = None,
        historic_future_covariates: Optional[TimeSeries] = None,
        future_covariates: Optional[TimeSeries] = None,
        num_samples: int = 1,
        verbose: bool = False,
    ) -> TimeSeries:
        """Forecasts values for a certain number of time steps after the end of the series.
        TransferableFutureCovariatesLocalForecastingModel must implement the predict logic in this method.
        """
        pass

    def _predict_wrapper(
        self,
        n: int,
        series: TimeSeries,
        past_covariates: Optional[TimeSeries],
        future_covariates: Optional[TimeSeries],
        num_samples: int,
        verbose: bool = False,
    ) -> TimeSeries:
        return self.predict(
            n=n,
            series=series,
            future_covariates=future_covariates,
            num_samples=num_samples,
            verbose=verbose,
        )

    def _supports_non_retrainable_historical_forecasts(self) -> bool:
        return True

    @property
    def _supress_generate_predict_encoding(self) -> bool:
        return True



"""
Models
------
"""

from darts.logging import get_logger

logger = get_logger(__name__)

# Forecasting
from darts.models.forecasting.arima import ARIMA
from darts.models.forecasting.auto_arima import AutoARIMA
from darts.models.forecasting.baselines import (
    NaiveDrift,
    NaiveMean,
    NaiveMovingAverage,
    NaiveSeasonal,
)
from darts.models.forecasting.exponential_smoothing import ExponentialSmoothing
from darts.models.forecasting.fft import FFT
from darts.models.forecasting.kalman_forecaster import KalmanForecaster
from darts.models.forecasting.linear_regression_model import LinearRegressionModel
from darts.models.forecasting.random_forest import RandomForest
from darts.models.forecasting.regression_ensemble_model import RegressionEnsembleModel
from darts.models.forecasting.regression_model import RegressionModel
from darts.models.forecasting.tbats_model import BATS, TBATS
from darts.models.forecasting.theta import FourTheta, Theta
from darts.models.forecasting.varima import VARIMA

try:
    from darts.models.forecasting.block_rnn_model import BlockRNNModel
    from darts.models.forecasting.dlinear import DLinearModel
    from darts.models.forecasting.nbeats import NBEATSModel
    from darts.models.forecasting.nhits import NHiTSModel
    from darts.models.forecasting.nlinear import NLinearModel
    from darts.models.forecasting.rnn_model import RNNModel
    from darts.models.forecasting.tcn_model import TCNModel
    from darts.models.forecasting.tft_model import TFTModel
    from darts.models.forecasting.transformer_model import TransformerModel


except ModuleNotFoundError:
    logger.warning(
        "Support for Torch based models not available. "
        'To enable them, install "darts", "u8darts[torch]" or "u8darts[all]" (with pip); '
        'or "u8darts-torch" or "u8darts-all" (with conda).'
    )

try:
    from darts.models.forecasting.lgbm import LightGBMModel
except ModuleNotFoundError:
    logger.warning(
        "Support for LightGBM not available. "
        "To enable LightGBM support in Darts, follow the detailed "
        "install instructions for LightGBM in the README: "
        "https://github.com/unit8co/darts/blob/master/INSTALL.md"
    )
    # TODO: simpler option would be to write LightGBMModel=None
    # Prevent a second ImportError that would interrupt the import
    class NotImportedLightGBM:
        usable = False

    LightGBMModel = NotImportedLightGBM()

try:
    from darts.models.forecasting.prophet_model import Prophet
except ImportError:
    logger.warning(
        "The prophet module could not be imported. "
        "To enable support for Prophet model, follow "
        "the instruction in the README: "
        "https://github.com/unit8co/darts/blob/master/INSTALL.md"
    )

    class NotImportedProphet:
        usable = False

    Prophet = NotImportedProphet()

try:
    from darts.models.forecasting.catboost_model import CatBoostModel
except ModuleNotFoundError:
    logger.warning(
        "The catboost module could not be imported. "
        "To enable support for CatBoostModel model, "
        "follow the instruction in the README: "
        "https://github.com/unit8co/darts/blob/master/INSTALL.md"
    )

    class NotImportedCatBoostModel:
        usable = False

    CatBoostModel = NotImportedCatBoostModel()

try:
    from darts.models.forecasting.croston import Croston
    from darts.models.forecasting.sf_auto_arima import StatsForecastAutoARIMA
    from darts.models.forecasting.sf_auto_ces import StatsForecastAutoCES
    from darts.models.forecasting.sf_auto_ets import StatsForecastAutoETS
    from darts.models.forecasting.sf_auto_theta import StatsForecastAutoTheta

except ImportError:
    logger.warning(
        "The statsforecast module could not be imported. "
        "To enable support for the StatsForecastAutoARIMA, "
        "StatsForecastAutoETS and Croston models, please consider "
        "installing it."
    )

    class NotImportedStatsForecastAutoARIMA:
        usable = False

    StatsForecastAutoARIMA = NotImportedStatsForecastAutoARIMA()

    class NotImportedStatsForecastAutoETS:
        usable = False

    StatsForecastAutoETS = NotImportedStatsForecastAutoETS()

    class NotImportedCroston:
        usable = False

    Croston = NotImportedCroston()

try:
    from darts.models.forecasting.xgboost import XGBModel
except ImportError:
    logger.warning(
        "The xgboost module could not be imported. "
        "To enable support for XGBoost model, install the xgboost package."
    )

    class NotImportedXGBModel:
        usable = False

    XGBModel = NotImportedXGBModel()

from darts.models.filtering.gaussian_process_filter import GaussianProcessFilter
from darts.models.filtering.kalman_filter import KalmanFilter

# Filtering
from darts.models.filtering.moving_average_filter import MovingAverageFilter
from darts.models.forecasting.baselines import NaiveEnsembleModel

# Ensembling
from darts.models.forecasting.ensemble_model import EnsembleModel



from itertools import product

import numpy as np

from darts.tests.base_test_class import DartsBaseTestClass
from darts.utils.data.tabularization import strided_moving_window


class StridedMovingWindowTestCase(DartsBaseTestClass):

    """
    Tests `strided_moving_window` function defined in `darts.utils.data.tabularization`.
    """

    def test_strided_moving_windows_extracted_windows(self):
        """
        Tests that each of the windows extracted by `strided_moving_windows`
        is correct over a number of input parameter combinations.

        This is achieved by looping over each extracted window, and checking that the
        `i`th window corresponds to the the next `window_len` values found after
        `i * stride` (i.e. the index position at which the `i`th window should begin).
        """
        # Parameter combos to test:
        window_len_combos = (1, 2, 5)
        axis_combos = (0, 1, 2)
        stride_combos = (1, 2, 3)
        # Create a 'dummy input' with linearly increasing values:
        x_shape = (10, 8, 12)
        x = np.arange(np.prod(x_shape)).reshape(*x_shape)
        for (axis, stride, window_len) in product(
            axis_combos, stride_combos, window_len_combos
        ):
            windows = strided_moving_window(x, window_len, stride, axis)
            # Iterate over extracted windows:
            for i in range(windows.shape[axis]):
                # All of the extract windows are found along the `axis` dimension; shift
                # `axis` so that it now appears as the final dimension and then extract
                # `i`th window:
                window = np.moveaxis(windows, axis, -1)[:, :, :, i]
                # `i`th window should begin at `i * stride`:
                window_start_idx = i * stride
                # Window should include next `window_len` values after window start;
                # shift `axis` to last dimension then extract expected window:
                expected = np.moveaxis(x, axis, -1)[
                    :, :, window_start_idx : window_start_idx + window_len
                ]
                self.assertTrue(np.allclose(window, expected))

    def test_strided_moving_window_invalid_stride_error(self):
        """
        Checks that appropriate error is thrown when `stride` is set to
        a non-positive number and/or a non-`int` value.
        """
        x = np.arange(1)
        # `stride` isn't positive:
        with self.assertRaises(ValueError) as e:
            strided_moving_window(x, window_len=1, stride=0)
        self.assertEqual(
            ("`stride` must be a positive `int`."),
            str(e.exception),
        )
        # `stride` is `float`, not `int`:
        with self.assertRaises(ValueError) as e:
            strided_moving_window(x, window_len=1, stride=1.1)
        self.assertEqual(
            ("`stride` must be a positive `int`."),
            str(e.exception),
        )

    def test_strided_moving_window_negative_window_len_error(self):
        """
        Checks that appropriate error is thrown when `wendow_len`
        is set to a non-positive number and/or a non-`int` value.
        """
        x = np.arange(1)
        # `window_len` isn't positive:
        with self.assertRaises(ValueError) as e:
            strided_moving_window(x, window_len=0, stride=1)
        self.assertEqual(
            ("`window_len` must be a positive `int`."),
            str(e.exception),
        )
        # `window_len` is `float`, not `int`:
        with self.assertRaises(ValueError) as e:
            strided_moving_window(x, window_len=1.1, stride=1)
        self.assertEqual(
            ("`window_len` must be a positive `int`."),
            str(e.exception),
        )

    def test_strided_moving_window_pass_invalid_axis_error(self):
        """
        Checks that appropriate error is thrown when `axis`
        is set to a non-`int` value, or a value not less than
        `x.ndim`.
        """
        x = np.arange(1)
        # `axis` NOT an int:
        with self.assertRaises(ValueError) as e:
            strided_moving_window(x, window_len=1, stride=1, axis=0.1)
        self.assertEqual(
            ("`axis` must be an `int` that is less than `x.ndim`."),
            str(e.exception),
        )
        # `axis` NOT < x.ndim:
        with self.assertRaises(ValueError) as e:
            strided_moving_window(x, window_len=1, stride=1, axis=1)
        self.assertEqual(
            ("`axis` must be an `int` that is less than `x.ndim`."),
            str(e.exception),
        )

    def test_strided_moving_window_window_len_too_large_error(self):
        """
        Checks that appropriate error is thrown when `window_len`
        is set to a value larger than `x.shape[axis]`.
        """
        x = np.arange(1)
        with self.assertRaises(ValueError) as e:
            strided_moving_window(x, window_len=2, stride=1)
        self.assertEqual(
            ("`window_len` must be less than or equal to x.shape[axis]."),
            str(e.exception),
        )



"""
NLL Gaussian Scorer
-------------------

Gaussian negative log-likelihood Scorer.

The anomaly score is the negative log likelihood of the actual series values
under a Gaussian distribution estimated from the stochastic predictions.
"""

import numpy as np
from scipy.stats import norm

from darts.ad.scorers.scorers import NLLScorer


class GaussianNLLScorer(NLLScorer):
    def __init__(self, window: int = 1) -> None:
        super().__init__(window=window)

    def __str__(self):
        return "GaussianNLLScorer"

    def _score_core_nllikelihood(
        self,
        deterministic_values: np.ndarray,
        probabilistic_estimations: np.ndarray,
    ) -> np.ndarray:

        mu = np.mean(probabilistic_estimations, axis=1)
        std = np.std(probabilistic_estimations, axis=1)
        return -norm.logpdf(deterministic_values, loc=mu, scale=std)



from numpy.random import RandomState

from darts.logging import get_logger
from darts.tests.base_test_class import DartsBaseTestClass

logger = get_logger(__name__)

try:
    import torch

    from darts.utils.torch import random_method

    TORCH_AVAILABLE = True
except ImportError:
    logger.warning("Torch not available. Torch utils will not be tested.")
    TORCH_AVAILABLE = False


if TORCH_AVAILABLE:
    # use a simple torch model mock
    class TorchModelMock:
        @random_method
        def __init__(self, some_params=None, **kwargs):
            self.model = torch.randn(5)
            # super().__init__()

        @random_method
        def fit(self, some_params=None):
            self.fit_value = torch.randn(5)

    class RandomMethodTestCase(DartsBaseTestClass):
        def test_it_raises_error_if_used_on_function(self):
            with self.assertRaises(ValueError):

                @random_method
                def a_random_function():
                    pass

        def test_model_is_random_by_default(self):
            model1 = TorchModelMock()
            model2 = TorchModelMock()
            self.assertFalse(torch.equal(model1.model, model2.model))

        def test_model_is_random_when_None_random_state_specified(self):
            model1 = TorchModelMock(random_state=None)
            model2 = TorchModelMock(random_state=None)
            self.assertFalse(torch.equal(model1.model, model2.model))

        def helper_test_reproducibility(self, model1, model2):
            self.assertTrue(torch.equal(model1.model, model2.model))

            model1.fit()
            model2.fit()
            self.assertTrue(torch.equal(model1.fit_value, model2.fit_value))

        def test_model_is_reproducible_when_seed_specified(self):
            model1 = TorchModelMock(random_state=42)
            model2 = TorchModelMock(random_state=42)
            self.helper_test_reproducibility(model1, model2)

        def test_model_is_reproducible_when_random_instance_specified(self):
            model1 = TorchModelMock(random_state=RandomState(42))
            model2 = TorchModelMock(random_state=RandomState(42))
            self.helper_test_reproducibility(model1, model2)

        def test_model_is_different_for_different_seeds(self):
            model1 = TorchModelMock(random_state=42)
            model2 = TorchModelMock(random_state=43)
            self.assertFalse(torch.equal(model1.model, model2.model))

        def test_model_is_different_for_different_random_instance(self):
            model1 = TorchModelMock(random_state=RandomState(42))
            model2 = TorchModelMock(random_state=RandomState(43))
            self.assertFalse(torch.equal(model1.model, model2.model))

        def helper_test_successive_call_are_different(self, model):
            # different between init and fit
            model.fit()
            self.assertFalse(torch.equal(model.model, model.fit_value))

            # different between 2 fit
            old_fit_value = model.fit_value.clone()
            model.fit()
            self.assertFalse(torch.equal(model.fit_value, old_fit_value))

        def test_successive_call_to_rng_are_different_when_seed_specified(self):
            model = TorchModelMock(random_state=42)
            self.helper_test_successive_call_are_different(model)

        def test_successive_call_to_rng_are_different_when_random_instance_specified(
            self,
        ):
            model = TorchModelMock(random_state=RandomState(42))
            self.helper_test_successive_call_are_different(model)

        def test_no_side_effect_between_rng_with_seeds(self):
            model = TorchModelMock(random_state=42)
            model.fit()
            fit_value = model.fit_value.clone()

            model = TorchModelMock(random_state=42)
            model2 = TorchModelMock(random_state=42)
            model2.fit()
            model.fit()

            self.assertTrue(torch.equal(model.fit_value, fit_value))

        def test_no_side_effect_between_rng_with_random_instance(self):
            model = TorchModelMock(random_state=RandomState(42))
            model.fit()
            fit_value = model.fit_value.clone()

            model = TorchModelMock(random_state=RandomState(42))
            model2 = TorchModelMock(random_state=RandomState(42))
            model2.fit()
            model.fit()

            self.assertTrue(torch.equal(model.fit_value, fit_value))



"""
PyODScorer
-----

This scorer can wrap around detection algorithms of PyOD.
`PyOD https://pyod.readthedocs.io/en/latest/#`_.
"""

from typing import Sequence

import numpy as np
from numpy.lib.stride_tricks import sliding_window_view
from pyod.models.base import BaseDetector

from darts.ad.scorers.scorers import FittableAnomalyScorer
from darts.logging import get_logger, raise_if_not
from darts.timeseries import TimeSeries

logger = get_logger(__name__)


class PyODScorer(FittableAnomalyScorer):
    def __init__(
        self,
        model: BaseDetector,
        window: int = 1,
        component_wise: bool = False,
        diff_fn="abs_diff",
    ) -> None:
        """
        When calling ``fit(series)``, a moving window is applied, which results in a set of vectors of size `W`,
        where `W` is the window size. The PyODScorer model is trained on these vectors. The ``score(series)``
        function will apply the same moving window and return the predicted raw anomaly score of each vector.

        Alternatively, the scorer has the functions ``fit_from_prediction()`` and ``score_from_prediction()``.
        Both require two series (actual and prediction), and compute a "difference" series by applying the
        function ``diff_fn`` (default: absolute difference). The resulting series is then passed to the
        functions ``fit()`` and ``score()``, respectively.

        `component_wise` is a boolean parameter indicating how the model should behave with multivariate inputs
        series. If set to True, the model will treat each series dimension independently by fitting a different
        PyODScorer model for each dimension. If set to False, the model concatenates the dimensions in
        each windows of length `W` and compute the score using only one underlying PyODScorer model.

        **Training with** ``fit()``:

        The input can be a series (univariate or multivariate) or multiple series. The series will be partitioned
        into equal size subsequences. The subsequence will be of size `W` * `D`, with:

        * `W` being the size of the window given as a parameter `window`
        * `D` being the dimension of the series (`D` = 1 if univariate or if `component_wise` is set to True)

        For a series of length `N`, (`N` - `W` + 1)/W subsequences will be generated. If a list of series is given
        of length L, each series will be partitioned into subsequences, and the results will be concatenated into
        an array of length L * number of subsequences of each series.

        The PyOD model will be fitted on the generated subsequences.

        If `component_wise` is set to True, the algorithm will be applied to each dimension independently. For each
        dimension, a PyOD model will be trained.

        **Computing score with** ``score()``:

        The input can be a series (univariate or multivariate) or a sequence of series. The given series must have the
        same dimension `D` as the data used to train the PyOD model.

        For each series, if the series is multivariate of dimension `D`:

        * if `component_wise` is set to False: it returns a univariate series (dimension=1). It represents
          the anomaly score of the entire series in the considered window at each timestamp.
        * if `component_wise` is set to True: it returns a multivariate series of dimension `D`. Each dimension
          represents the anomaly score of the corresponding component of the input.

        If the series is univariate, it returns a univariate series regardless of the parameter
        `component_wise`.

        A window of size `W` is rolled on the series with a stride equal to 1. It is the same size window `W` used
        during the training phase.
        Each value in the score series thus represents how anomalous the sample of the `W` previous values is.

        Parameters
        ----------
        model
            The (fitted) PyOD BaseDetector model.
        window
            Size of the window used to create the subsequences of the series.
        diff_fn
            Optionally, reduced function to use if two series are given. It will transform the two series into one.
            This allows the KMeansScorer to apply PyODScorer on the original series or on its residuals (difference
            between the prediction and the original series).
            Must be one of "abs_diff" and "diff" (defined in ``_diff_series()``).
            Default: "abs_diff"
        component_wise
            Boolean value indicating if the score needs to be computed for each component independently (True)
            or by concatenating the component in the considered window to compute one score (False).
            Default: False
        """

        raise_if_not(
            isinstance(model, BaseDetector),
            f"model must be a PyOD BaseDetector, found type: {type(model)}",
        )
        self.model = model

        raise_if_not(
            type(component_wise) is bool,
            f"Parameter `component_wise` must be Boolean, found type: {type(component_wise)}.",
        )
        self.component_wise = component_wise

        super().__init__(
            univariate_scorer=(not component_wise), window=window, diff_fn=diff_fn
        )

    def __str__(self):
        return "PyODScorer (model {})".format(self.model.__str__().split("(")[0])

    def _fit_core(self, list_series: Sequence[TimeSeries]):

        list_np_series = [series.all_values(copy=False) for series in list_series]

        # TODO: can we factorize code in common bteween PyODScorer and KMeansScorer?

        if not self.component_wise:
            self.model.fit(
                np.concatenate(
                    [
                        sliding_window_view(ar, window_shape=self.window, axis=0)
                        .transpose(0, 3, 1, 2)
                        .reshape(-1, self.window * len(ar[0]))
                        for ar in list_np_series
                    ]
                )
            )
        else:
            models = []
            for component_idx in range(self.width_trained_on):

                model_width = self.model
                model_width.fit(
                    np.concatenate(
                        [
                            sliding_window_view(
                                ar[:, component_idx], window_shape=self.window, axis=0
                            )
                            .transpose(0, 2, 1)
                            .reshape(-1, self.window)
                            for ar in list_np_series
                        ]
                    )
                )
                models.append(model_width)
            self.models = models

    def _score_core(self, series: TimeSeries) -> TimeSeries:

        raise_if_not(
            self.width_trained_on == series.width,
            "Input must have the same number of components as the data used for training"
            + " the PyODScorer model {},".format(self.model.__str__().split("(")[0])
            + f" found number of components equal to {series.width} and expected "
            + f"{self.width_trained_on}.",
        )

        np_series = series.all_values(copy=False)
        np_anomaly_score = []

        if not self.component_wise:

            np_anomaly_score.append(
                self.model.decision_function(
                    sliding_window_view(np_series, window_shape=self.window, axis=0)
                    .transpose(0, 3, 1, 2)
                    .reshape(-1, self.window * series.width)
                )
            )
        else:

            for component_idx in range(self.width_trained_on):
                score = self.models[component_idx].decision_function(
                    sliding_window_view(
                        np_series[:, component_idx],
                        window_shape=self.window,
                        axis=0,
                    )
                    .transpose(0, 2, 1)
                    .reshape(-1, self.window)
                )

                np_anomaly_score.append(score)

        return TimeSeries.from_times_and_values(
            series.time_index[self.window - 1 :], list(zip(*np_anomaly_score))
        )



"""
Missing Values Filler
---------------------
"""

from typing import Any, Mapping, Union

from darts import TimeSeries
from darts.logging import get_logger, raise_if, raise_if_not
from darts.utils.missing_values import fill_missing_values

from .base_data_transformer import BaseDataTransformer

logger = get_logger(__name__)


class MissingValuesFiller(BaseDataTransformer):
    def __init__(
        self,
        fill: Union[str, float] = "auto",
        name: str = "MissingValuesFiller",
        n_jobs: int = 1,
        verbose: bool = False,
    ):
        """Data transformer to fill missing values from a (sequence of) deterministic ``TimeSeries``.

        Parameters
        ----------
        fill
            The value used to replace the missing values.
            If set to 'auto', will auto-fill missing values using the :func:`pd.Dataframe.interpolate()` method.
        name
            A specific name for the transformer
        n_jobs
            The number of jobs to run in parallel. Parallel jobs are created only when a ``Sequence[TimeSeries]`` is
            passed as input to a method, parallelising operations regarding different ``TimeSeries``. Defaults to `1`
            (sequential). Setting the parameter to `-1` means using all the available processors.
            Note: for a small amount of data, the parallelisation overhead could end up increasing the total
            required amount of time.
        verbose
            Optionally, whether to print operations progress

        Examples
        --------
        >>> import numpy as np
        >>> from darts import TimeSeries
        >>> from darts.dataprocessing.transformers import MissingValuesFiller
        >>> values = np.arange(start=0, stop=1, step=0.1)
        >>> values[5:8] = np.nan
        >>> series = TimeSeries.from_values(values)
        >>> transformer = MissingValuesFiller()
        >>> series_filled = transformer.transform(series)
        >>> print(series_filled)
        <TimeSeries (DataArray) (time: 10, component: 1, sample: 1)>
        array([[[0. ]],
            [[0.1]],
            [[0.2]],
            [[0.3]],
            [[0.4]],
            [[0.5]],
            [[0.6]],
            [[0.7]],
            [[0.8]],
            [[0.9]]])
        Coordinates:
        * time       (time) int64 0 1 2 3 4 5 6 7 8 9
        * component  (component) object '0'
        Dimensions without coordinates: sample
        """
        raise_if_not(
            isinstance(fill, str) or isinstance(fill, float),
            "`fill` should either be a string or a float",
            logger,
        )
        raise_if(
            isinstance(fill, str) and fill != "auto",
            "invalid string for `fill`: can only be set to 'auto'",
            logger,
        )
        # Define fixed params (i.e. attributes defined before calling `super().__init__`):
        self._fill = fill
        super().__init__(name=name, n_jobs=n_jobs, verbose=verbose)

    @staticmethod
    def ts_transform(
        series: TimeSeries, params: Mapping[str, Any], **kwargs
    ) -> TimeSeries:
        return fill_missing_values(series, params["fixed"]["_fill"], **kwargs)
